# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import getopt
import shlex
import sys
import time
from collections import namedtuple
from string import ascii_uppercase

import urwid

from topydo.Commands import get_subcommand
from topydo.lib import TodoList
from topydo.lib.Config import ConfigError, config
from topydo.lib.Filter import (DependencyFilter, HiddenTagFilter,
                               RelevanceFilter, get_filter_list)
from topydo.lib.Sorter import Sorter
from topydo.lib.TodoFileWatched import TodoFileWatched
from topydo.lib.Utils import get_terminal_size
from topydo.lib.View import View
from topydo.ui.CLIApplicationBase import (GENERIC_HELP, CLIApplicationBase,
                                          error)
from topydo.ui.columns.ColumnCompleter import ColumnCompleter
from topydo.ui.columns.ColumnLayout import columns
from topydo.ui.columns.CommandLineWidget import CommandLineWidget
from topydo.ui.columns.ConsoleWidget import ConsoleWidget
from topydo.ui.columns.KeystateWidget import KeystateWidget
from topydo.ui.columns.TodoListWidget import TodoListWidget
from topydo.ui.columns.TodoWidget import TodoWidget
from topydo.ui.columns.Transaction import Transaction
from topydo.ui.columns.Utils import PaletteItem, to_urwid_color
from topydo.ui.columns.ViewWidget import ViewWidget


class UIView(View):
    """
    A subclass of view holding user input data that constructed the view (i.e.
    the sort expression and the filter expression, etc.)
    """
    def __init__(self, p_sorter, p_filter, p_todolist, p_data):
        super().__init__(p_sorter, p_filter, p_todolist)
        self.data = p_data

_APPEND_COLUMN = 1
_EDIT_COLUMN = 2
_COPY_COLUMN = 3
_INSERT_COLUMN = 4


class CliWrapper(urwid.Pile):
    """
    Simple wrapper widget for CommandLineWidget with KeystateWidget and
    CompletionBoxWidget rendered dynamically.
    """
    def __init__(self, *args, **kwargs):
        self.width = 0

        super().__init__(*args, **kwargs)

    def render(self, p_size, focus):
        self.width = p_size[0]
        return super().render(p_size, focus)


class MainPile(urwid.Pile):
    """
    This subclass of Pile doesn't change focus on cursor up/down / mouse press
    events. The implementation was taken from its base class.
    """
    def __init__(self, p_widget_list, p_focus_item=None):
        urwid.register_signal(MainPile, ['blur_console'])

        super().__init__(p_widget_list, p_focus_item)

    def mouse_event(self, p_size, p_event, p_button, p_col, p_row, p_focus):
        if self.focus_position != 2:
            urwid.emit_signal(self, 'blur_console')

        return super().mouse_event(p_size, p_event, p_button, p_col, p_row, p_focus)  # pylint: disable=E1102

    def keypress(self, p_size, p_key):
        if not self.contents:
            return p_key

        item_rows = None
        if len(p_size) == 2:
            item_rows = self.get_item_rows(p_size, focus=True)

        i = self.focus_position
        if self.selectable():
            tsize = self.get_item_size(p_size, i, True, item_rows)
            key = self.focus.keypress(tsize, p_key)
            if self._command_map[key] not in ('cursor up', 'cursor down'):
                return key


class UIApplication(CLIApplicationBase):
    def __init__(self):
        super().__init__()

        args = self._process_flags()

        try:
            opts, args = getopt.getopt(args[1:], 'l:')
        except getopt.GetoptError as e:
            error(str(e))
            sys.exit(1)

        self.alt_layout_path = None

        for opt, value in opts:
            if opt == "-l":
                self.alt_layout_path = value

        def callback():
            self.todolist.erase()
            self.todolist.add_list(self.todofile.read())
            self._update_all_columns()
            self._redraw()

        self.column_width = config().column_width()
        self.todofile = TodoFileWatched(config().todotxt(), callback)
        self.todolist = TodoList.TodoList(self.todofile.read())

        self.marked_todos = set()

        self.columns = urwid.Columns([], dividechars=0,
                                     min_width=config().column_width())
        self.columns.contents.set_focus_changed_callback(self._move_highlight)
        completer = ColumnCompleter(self.todolist)
        self.commandline = CommandLineWidget(completer, 'topydo> ')
        self.keystate_widget = KeystateWidget()
        self.status_line = urwid.Columns([
            ('weight', 1, urwid.Filler(self.commandline)),
        ])
        self.cli_wrapper = CliWrapper([(1, self.status_line)])

        self.keymap = config().column_keymap()
        self._alarm = None

        self._last_cmd = None

        # console widget
        self.console = ConsoleWidget()
        get_terminal_size(self._console_width)

        urwid.connect_signal(self.commandline, 'blur', self._blur_commandline)
        urwid.connect_signal(self.commandline, 'execute_command',
                             self._execute_handler)
        urwid.connect_signal(self.commandline, 'show_completions', self._show_completion_box)
        urwid.connect_signal(self.commandline, 'hide_completions', self._hide_completion_box)

        def hide_console(p_focus_commandline=False):
            if p_focus_commandline:
                self._focus_commandline()
            else:
                self._console_visible = False
        urwid.connect_signal(self.console, 'close', hide_console)

        # view widget
        self.viewwidget = ViewWidget(self.todolist)

        urwid.connect_signal(self.viewwidget, 'save',
                             lambda: self._update_view(self.viewwidget.data))

        def hide_viewwidget():
            # prevent the view widget to be hidden when the last column was
            # deleted
            if self.columns.contents:
                self._viewwidget_visible = False
                self._blur_commandline()

        urwid.connect_signal(self.viewwidget, 'close', hide_viewwidget)

        self.mainwindow = MainPile([
            ('weight', 1, self.columns),
            ('pack', self.cli_wrapper),
        ])

        urwid.connect_signal(self.mainwindow, 'blur_console', hide_console)

        # the columns should have keyboard focus
        self._blur_commandline()

        self._screen = urwid.raw_display.Screen()

        def create_color_palette():
            project_color = to_urwid_color(config().project_color())
            context_color = to_urwid_color(config().context_color())
            metadata_color = to_urwid_color(config().metadata_color())
            link_color = to_urwid_color(config().link_color())
            focus_background_color = to_urwid_color(config().focus_background_color())
            marked_background_color = to_urwid_color(config().marked_background_color())

            palette = [
                (PaletteItem.PROJECT, '', '', '', project_color, ''),
                (PaletteItem.PROJECT_FOCUS, '', 'light gray', '', project_color, focus_background_color),
                (PaletteItem.CONTEXT, '', '', '', context_color, ''),
                (PaletteItem.CONTEXT_FOCUS, '', 'light gray', '', context_color, focus_background_color),
                (PaletteItem.METADATA, '', '', '', metadata_color, ''),
                (PaletteItem.METADATA_FOCUS, '', 'light gray', '', metadata_color, focus_background_color),
                (PaletteItem.LINK, '', '', '', link_color, ''),
                (PaletteItem.LINK_FOCUS, '', 'light gray', '', link_color, focus_background_color),
                (PaletteItem.DEFAULT_FOCUS, '', 'light gray', '', '', focus_background_color),
                (PaletteItem.MARKED, '', 'light blue', '', '', marked_background_color),
            ]

            for C in ascii_uppercase:
                pri_color_cfg = config().priority_color(C)

                pri_color = to_urwid_color(pri_color_cfg)
                pri_color_focus = pri_color if not pri_color_cfg.is_neutral() else 'black'

                palette.append((
                    'pri_' + C, '', '', '', pri_color, ''
                ))
                palette.append((
                    'pri_' + C + '_focus', '', 'light gray', '', pri_color_focus, focus_background_color
                ))

            return palette

        def create_mono_palette():
            palette = [
                (PaletteItem.DEFAULT_FOCUS, 'black', 'light gray'),
                (PaletteItem.PROJECT_FOCUS, PaletteItem.DEFAULT_FOCUS),
                (PaletteItem.CONTEXT_FOCUS, PaletteItem.DEFAULT_FOCUS),
                (PaletteItem.METADATA_FOCUS, PaletteItem.DEFAULT_FOCUS),
                (PaletteItem.LINK_FOCUS, PaletteItem.DEFAULT_FOCUS),
                (PaletteItem.MARKED, 'default,underline,bold', 'default'),
            ]

            for C in ascii_uppercase:
                palette.append(
                    ('pri_' + C + '_focus', PaletteItem.DEFAULT_FOCUS)
                )

            return palette

        if config().colors():
            self._screen.register_palette(create_color_palette())
        else:
            self._screen.register_palette(create_mono_palette())

        self._screen.set_terminal_properties(256)

        self.mainloop = urwid.MainLoop(
            self.mainwindow,
            screen=self._screen,
            unhandled_input=self._handle_input,
            pop_ups=True
        )

        self.column_mode = _APPEND_COLUMN
        self._set_alarm_for_next_midnight_update()

    def _move_highlight(self, p_new_focus):
        """
        Removes highlight from currently focused column and applies it on
        column with index equal to p_new_focus.
        """
        self.columns.focus.highlight(False)
        self.columns.contents[p_new_focus][0].highlight(True)
        self.columns._invalidate()

    def _set_alarm_for_next_midnight_update(self):
        def callback(p_loop, p_data):
            TodoWidget.wipe_cache()
            self._update_all_columns()
            self._set_alarm_for_next_midnight_update()

        tomorrow = datetime.datetime.now() + datetime.timedelta(days=1)
        # turn it into midnight
        tomorrow = tomorrow.replace(hour=0, minute=0, second=0, microsecond=0)

        self.mainloop.set_alarm_at(time.mktime(tomorrow.timetuple()), callback)

    def _output(self, p_text):
        self._print_to_console(p_text)

    def _check_id_validity(self, p_ids):
        """
        Checks if there are any invalid todo IDs in p_ids list.

        Returns proper error message if any ID is invalid and None otherwise.
        """
        errors = []
        valid_ids = self.todolist.ids()

        if len(p_ids) == 0:
            errors.append('No todo item was selected')
        else:
            errors = ["Invalid todo ID: {}".format(todo_id)
                      for todo_id in p_ids - valid_ids]

        errors = '\n'.join(errors) if errors else None
        return errors

    def _execute_handler(self, p_command, p_todo_id=None, p_output=None):
        """
        Executes a command, given as a string.
        """
        p_output = p_output or self._output
        self._console_visible = False

        self._last_cmd = (p_command, p_output == self._output)

        try:
            p_command = shlex.split(p_command)
        except ValueError as verr:
            self._print_to_console('Error: ' + str(verr))
            return

        try:
            subcommand, args = get_subcommand(p_command)
        except ConfigError as cerr:
            self._print_to_console(
                'Error: {}. Check your aliases configuration.'.format(cerr))
            return

        if subcommand is None:
            self._print_to_console(GENERIC_HELP)
            return

        env_args = (self.todolist, p_output, self._output, self._input)
        ids = None

        if '{}' in args:
            if self._has_marked_todos():
                ids = self.marked_todos
            else:
                ids = {p_todo_id} if p_todo_id else set()

            invalid_ids = self._check_id_validity(ids)

            if invalid_ids:
                self._print_to_console('Error: ' + invalid_ids)
                return

        transaction = Transaction(subcommand, env_args, ids)
        transaction.prepare(args)
        label = transaction.label
        self._backup(subcommand, p_label=label)

        try:
            if transaction.execute():
                post_archive_action = transaction.execute_post_archive_actions
                self._post_archive_action = post_archive_action
                self._post_execute()
            else:
                self._rollback()
        except TypeError:
            # TODO: show error message
            pass

    def _update_all_columns(self):
        for column, _ in self.columns.contents:
            column.update()
            column.keystate = None

    def _post_execute(self):
        # store dirty flag because base _post_execute will reset it after flush
        dirty = self.todolist.dirty
        super()._post_execute()

        if dirty or self.marked_todos:
            self._reset_state()

    def _rollback(self):
        try:
            self.backup.apply(self.todolist, p_archive=None)
        except AttributeError:
            pass

    def _repeat_last_cmd(self, p_todo_id=None):
        try:
            cmd, verbosity = self._last_cmd
        except TypeError:
            return

        self._execute_handler(cmd, p_todo_id,
                              self._output if verbosity else lambda _: None)

    def _reset_state(self):
        for widget in TodoWidget.cache.values():
            widget.unmark()
        self.marked_todos.clear()
        self._update_all_columns()

    def _blur_commandline(self):
        self._console_visible = False
        self.mainwindow.focus_item = 0

    def _focus_commandline(self):
        self.mainwindow.focus_item = 1

    def _focus_first_column(self):
        self.columns.focus_position = 0

    def _focus_last_column(self):
        end_pos = len(self.columns.contents) - 1
        self.columns.focus_position = end_pos

    def _focus_next_column(self):
        size = len(self.columns.contents)
        if self.columns.focus_position < size -1:
            self.columns.focus_position += 1

    def _focus_previous_column(self):
        if self.columns.focus_position > 0:
            self.columns.focus_position -= 1

    def _append_column(self):
        self.viewwidget.reset()
        self.column_mode = _APPEND_COLUMN
        self._viewwidget_visible = True

    def _insert_column(self):
        self.viewwidget.reset()
        self.column_mode = _INSERT_COLUMN
        self._viewwidget_visible = True

    def _edit_column(self):
        self.viewwidget.data = self.columns.focus.view.data
        self.column_mode = _EDIT_COLUMN
        self._viewwidget_visible = True

    def _delete_column(self):
        try:
            focus = self.columns.focus_position
            del self.columns.contents[focus]

            if self.columns.contents:
                self.columns.focus_position = focus
            else:
                self._append_column()
        except IndexError:
            # no columns
            pass

    def _copy_column(self):
        self.viewwidget.data = self.columns.focus.view.data
        self.column_mode = _COPY_COLUMN
        self._viewwidget_visible = True

    def _column_action_handler(self, p_action):
        dispatch = {
            'first_column': self._focus_first_column,
            'last_column': self._focus_last_column,
            'prev_column': self._focus_previous_column,
            'next_column': self._focus_next_column,
            'append_column': self._append_column,
            'insert_column': self._insert_column,
            'edit_column': self._edit_column,
            'delete_column': self._delete_column,
            'copy_column': self._copy_column,
            'swap_left': self._swap_column_left,
            'swap_right': self._swap_column_right,
            'reset': self._reset_state,
        }
        dispatch[p_action]()

    def _handle_input(self, p_input):
        dispatch = {
            ':': self._focus_commandline,
        }

        try:
            dispatch[p_input]()
        except KeyError:
            # the key is unknown, ignore
            pass

    def _viewdata_to_view(self, p_data):
        """
        Converts a dictionary describing a view to an actual UIView instance.
        """
        sorter = Sorter(p_data['sortexpr'], p_data['groupexpr'])
        filters = []

        if not p_data['show_all']:
            filters.append(DependencyFilter(self.todolist))
            filters.append(RelevanceFilter())
            filters.append(HiddenTagFilter())

        filters += get_filter_list(p_data['filterexpr'].split())

        return UIView(sorter, filters, self.todolist, p_data)

    def _update_view(self, p_data):
        """ Creates a view from the data entered in the view widget. """
        view = self._viewdata_to_view(p_data)

        if self.column_mode == _APPEND_COLUMN or self.column_mode == _COPY_COLUMN:
            self._add_column(view)
        elif self.column_mode == _INSERT_COLUMN:
            self._add_column(view, self.columns.focus_position)
        elif self.column_mode == _EDIT_COLUMN:
            current_column = self.columns.focus

            current_column.title = p_data['title']
            current_column.view = view

        self._viewwidget_visible = False
        self._blur_commandline()

    def _add_column(self, p_view, p_pos=None):
        """
        Given an UIView, adds a new column widget with the todos in that view.

        When no position is given, it is added to the end, otherwise inserted
        before that position.
        """
        def execute_silent(p_cmd, p_todo_id=None):
            self._execute_handler(p_cmd, p_todo_id, lambda _: None)

        todolist = TodoListWidget(p_view, p_view.data['title'], self.keymap)
        urwid.connect_signal(todolist, 'execute_command_silent',
                             execute_silent)
        urwid.connect_signal(todolist, 'execute_command', self._execute_handler)
        urwid.connect_signal(todolist, 'repeat_cmd', self._repeat_last_cmd)
        urwid.connect_signal(todolist, 'refresh', self.mainloop.screen.clear)
        urwid.connect_signal(todolist, 'add_pending_action', self._set_alarm)
        urwid.connect_signal(todolist, 'remove_pending_action', self._remove_alarm)
        urwid.connect_signal(todolist, 'column_action', self._column_action_handler)
        urwid.connect_signal(todolist, 'show_keystate', self._print_keystate)
        urwid.connect_signal(todolist, 'toggle_mark',
                             self._process_mark_toggle)

        options = self.columns.options(
            width_type='given',
            width_amount=config().column_width(),
            box_widget=True
        )

        item = (todolist, options)

        if p_pos == None:
            p_pos = len(self.columns.contents)

        self.columns.contents.insert(p_pos, item)

        self.columns.focus_position = p_pos
        self._blur_commandline()

    def _print_keystate(self, p_keystate):
        self.keystate_widget.set_text(p_keystate)
        self._keystate_visible = len(p_keystate) > 0

    def _show_completion_box(self):
        contents = self.cli_wrapper.contents
        if len(contents) == 1:
            completion_box = self.commandline.completion_box
            opts = ('given', completion_box.height)

            max_width = self.cli_wrapper.width
            pos = self.commandline.get_cursor_coords((max_width,))[0]
            l_margin = pos - completion_box.margin
            r_margin = max_width - pos - completion_box.min_width + completion_box.margin

            padding = urwid.Padding(completion_box,
                                    min_width=completion_box.min_width,
                                    left=l_margin,
                                    right=r_margin)

            contents.insert(0, (padding, opts))
            completion_box.focus.set_attr_map({None: PaletteItem.MARKED})
            self.cli_wrapper.focus_position = 1

    def _hide_completion_box(self):
        contents = self.cli_wrapper.contents
        if len(contents) == 2:
            del contents[0]

        self.cli_wrapper.focus_position = 0

    def _set_alarm(self, p_callback):
        """ Sets alarm to execute p_action specified in 0.5 sec. """
        self._alarm = self.mainloop.set_alarm_in(0.5, p_callback)

    def _remove_alarm(self):
        """ Removes pending action alarm stored in _alarm attribute. """
        self.mainloop.remove_alarm(self._alarm)
        self._alarm = None

    def _swap_column_left(self):
        pos = self.columns.focus_position
        if pos > 0:
            _columns = self.columns.contents
            _columns[pos], _columns[pos - 1] = _columns[pos - 1], _columns[pos]
            self.columns.focus_position -= 1

    def _swap_column_right(self):
        pos = self.columns.focus_position
        _columns = self.columns.contents
        if pos < len(_columns) - 1:
            _columns[pos], _columns[pos + 1] = _columns[pos + 1], _columns[pos]
            self.columns.focus_position += 1

    @property
    def _console_visible(self):
        contents = self.mainwindow.contents
        return len(contents) == 3 and isinstance(contents[2][0], ConsoleWidget)

    @_console_visible.setter
    def _console_visible(self, p_enabled):
        contents = self.mainwindow.contents

        if p_enabled == True and len(contents) == 2:
            contents.append((self.console, ('pack', None)))
            self.mainwindow.focus_position = 2
        elif p_enabled == False and self._console_visible:
            self.console.clear()
            del contents[2]
            self.mainwindow.focus_position = 0

    @property
    def _keystate_visible(self):
        contents = self.status_line.contents
        return len(contents) == 2 and isinstance(contents[1][0].original_widget,
                                                 KeystateWidget)

    @_keystate_visible.setter
    def _keystate_visible(self, p_enabled):
        contents = self.status_line.contents

        if p_enabled and len(contents) == 1:
            contents.append((urwid.Filler(self.keystate_widget),
                             ('weight', 1, True)))
        elif not p_enabled and self._keystate_visible:
            del contents[1]

    @property
    def _viewwidget_visible(self):
        contents = self.mainwindow.contents
        return len(contents) == 3 and isinstance(contents[2][0], ViewWidget)

    @_viewwidget_visible.setter
    def _viewwidget_visible(self, p_enabled):
        contents = self.mainwindow.contents

        if p_enabled == True and len(contents) == 2:
            contents.append((self.viewwidget, ('pack', None)))
            self.mainwindow.focus_position = 2
        elif p_enabled == False and self._viewwidget_visible:
            del contents[2]

    def _print_to_console(self, p_text):
        self._console_visible = True
        self.console.print_text(p_text)

    def _redraw(self):
        self.mainloop.draw_screen()

    def _input(self, p_question):
        self._print_to_console(p_question)

        # don't wait for the event loop to enter idle, there is a command
        # waiting for input right now, so already go ahead and draw the
        # question on screen.
        self._redraw()

        user_input = self.mainloop.screen.get_input()
        self._console_visible = False

        return user_input[0]

    def _console_width(self):
        terminal_size = namedtuple('Terminal_Size', 'columns lines')
        width = self.cli_wrapper.width - 2
        sz = terminal_size(width, 1)

        return sz

    def _has_marked_todos(self):
        return len(self.marked_todos) > 0

    def _process_mark_toggle(self, p_todo_id, p_force=None):
        """
        Adds p_todo_id to marked_todos attribute and returns True if p_todo_id
        is not already marked. Removes p_todo_id from marked_todos and returns
        False otherwise.

        p_force parameter accepting 'mark' or 'unmark' values, if set, can force
        desired action without checking p_todo_id presence in marked_todos.
        """
        if p_force in ['mark', 'unmark']:
            action = p_force
        else:
            action = 'mark' if p_todo_id not in self.marked_todos else 'unmark'

        if action == 'mark':
            self.marked_todos.add(p_todo_id)
            return True
        else:
            self.marked_todos.remove(p_todo_id)
            return False

    def run(self):
        layout = columns(self.alt_layout_path)
        if len(layout) > 0:
            for column in layout:
                self._add_column(self._viewdata_to_view(column))
        else:
            dummy = {
                "title": "All tasks",
                "sortexpr": "desc:prio",
                "groupexpr": "",
                "filterexpr": "",
                "show_all": True,
            }
            self._add_column(self._viewdata_to_view(dummy))

        # make sure that the first column is focused on startup
        self.columns.focus_position = 0

        while True:
            try:
                self.mainloop.run()
            except KeyboardInterrupt:
                self._print_to_console("Use the 'quit' command to exit topydo.")

if __name__ == '__main__':
    UIApplication().run()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from topydo.lib.Color import AbstractColor
from topydo.lib.Todo import Todo
from topydo.lib.TopydoString import TopydoString
from topydo.ui.columns.Utils import PaletteItem

PALETTE_LOOKUP = {
    # omitting AbstractColor.NEUTRAL on purpose, so a text without any
    # attribute will be added to the markup
    AbstractColor.PROJECT: PaletteItem.PROJECT,
    AbstractColor.CONTEXT: PaletteItem.CONTEXT,
    AbstractColor.META: PaletteItem.METADATA,
    AbstractColor.LINK: PaletteItem.LINK,
}

def topydostringToMarkup(p_string):
    markup = []

    color_positions = sorted(p_string.colors.items())

    # in case no color positions are available, at least set something at the
    # start position
    if not color_positions:
        color_positions = [(0, None)]

    for i, (start_pos, color) in enumerate(color_positions):
        # color starts at indicated position
        start = start_pos

        # color ends at next color indication. if missing, run until the end of
        # the string
        try:
            end = color_positions[i+1][0]
        except IndexError:
            end = len(str(p_string))

        text = str(p_string)[start:end]

        if color in PALETTE_LOOKUP:
            markup.append((PALETTE_LOOKUP[color], text))
        else:
            # a plain text without any attribute set (including
            # AbstractColor.NEUTRAL)
            markup.append(text)

    color_at_start = color_positions and color_positions[0][0] == 0

    # priority color should appear at the start if present, build a nesting
    # markup
    if color_at_start and isinstance(p_string.metadata, Todo):
        priority = p_string.metadata.priority()

        if priority:
            markup = ('pri_' + priority, markup)

    return markup

class ConsoleWidget(urwid.LineBox):
    def __init__(self, p_text=""):
        urwid.register_signal(ConsoleWidget, ['close'])

        self.width = 0
        self.pile = urwid.Pile([])

        super().__init__(self.pile)

    def keypress(self, p_size, p_key):
        if p_key == 'enter' or p_key == 'q' or p_key == 'esc':
            urwid.emit_signal(self, 'close')

        # don't return the key, 'enter', 'escape', 'q' or ':' are your only
        # escape. ':' will reenter to the cmdline.
        elif p_key == ':':
            urwid.emit_signal(self, 'close', True)

    def selectable(self):
        return True

    def print_text(self, p_text):
        if isinstance(p_text, list):
            for text in p_text:
                self.print_text(text)

            return
        elif isinstance(p_text, TopydoString):
            text = urwid.Text(topydostringToMarkup(p_text))
        else:
            text = urwid.Text(p_text)

        self.pile.contents.append((text, ('pack', None)))

    def clear(self):
        self.pile.contents = []



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Entry file for the topydo Prompt interface (CLI). """

import shlex
import sys

from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.shortcuts import prompt

from topydo.Commands import get_subcommand
from topydo.lib.Config import ConfigError, config
from topydo.lib.TodoFileWatched import TodoFileWatched
from topydo.ui.CLIApplicationBase import (GENERIC_HELP, CLIApplicationBase,
                                          error)
from topydo.ui.prompt.PromptCompleter import PromptCompleter

# First thing is to poke the configuration and check whether it's sane
# The modules below may already read in configuration upon import, so
# make sure to bail out if the configuration is invalid.
try:
    config()
except ConfigError as config_error:
    error(str(config_error))
    sys.exit(1)



class PromptApplication(CLIApplicationBase):
    """
    This class implements a variant of topydo's CLI showing a shell and
    offering auto-completion thanks to the prompt toolkit.
    """

    def __init__(self):
        super().__init__()

        self._process_flags()
        self.completer = None
        self.todofile = TodoFileWatched(config().todotxt(), self._load_file)

    def _load_file(self):
        """
        Reads the configured todo.txt file and loads it into the todo list
        instance.
        """
        self.todolist.erase()
        self.todolist.add_list(self.todofile.read())
        self.completer = PromptCompleter(self.todolist)

    def run(self):
        """ Main entry function. """
        history = InMemoryHistory()
        self._load_file()

        while True:
            # (re)load the todo.txt file (only if it has been modified)

            try:
                user_input = prompt(u'topydo> ', history=history,
                                    completer=self.completer,
                                    complete_while_typing=False)
                user_input = shlex.split(user_input)
            except EOFError:
                sys.exit(0)
            except KeyboardInterrupt:
                continue
            except ValueError as verr:
                error('Error: ' + str(verr))
                continue

            try:
                (subcommand, args) = get_subcommand(user_input)
            except ConfigError as ce:
                error('Error: ' + str(ce) + '. Check your aliases configuration')
                continue

            try:
                if self._execute(subcommand, args) != False:
                    self._post_execute()
            except TypeError:
                print(GENERIC_HELP)


def main():
    """ Main entry point of the prompt interface. """
    PromptApplication().run()

if __name__ == '__main__':
    main()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module serves for managing todo and archive changesets. """

import json
import time
import zlib
from copy import deepcopy
from hashlib import sha1
from os import path

from topydo.lib.Config import config
from topydo.lib.TodoList import TodoList


def hash_todolist(p_todolist):
    """ Calculates hash for TodoList.TodoList object. """
    todolist_hash = sha1(p_todolist.print_todos().encode('utf-8')).hexdigest()

    return todolist_hash

def get_backup_path():
    """ Returns full path and filename of backup file """
    dirname, filename = path.split(path.splitext(config().todotxt())[0])
    filename = '.' + filename + '.bak'

    return path.join(dirname, filename)

class ChangeSet(object):
    """ Class for operations related with backup management. """

    def __init__(self, p_todolist=None, p_archive=None, p_label=None):
        self.todolist = deepcopy(p_todolist)
        self.archive = deepcopy(p_archive)
        self.timestamp = str(time.time())
        self.label = ' '.join(p_label if p_label else [])

        try:
            self.json_file = open(get_backup_path(), 'r+b')
        except IOError:
            self.json_file = open(get_backup_path(), 'w+b')

        self._read()

    def __iter__(self):
        items = {key: self.backup_dict[key]
                 for key in self.backup_dict if key != 'index'}.items()
        return iter(sorted(items, reverse=True))

    def _read(self):
        """
        Reads backup file from json_file property and sets backup_dict property
        with data decompressed and deserialized from that file. If no usable
        data is found backup_dict is set to the empty dict.
        """
        self.json_file.seek(0)
        try:
            data = zlib.decompress(self.json_file.read())
            self.backup_dict = json.loads(data.decode('utf-8'))
        except (EOFError, zlib.error):
            self.backup_dict = {}

    def _write(self):
        """
        Writes data from backup_dict property in serialized and compressed form
        to backup file pointed in json_file property.
        """
        self.json_file.seek(0)
        self.json_file.truncate()
        dump = json.dumps(self.backup_dict)
        dump_c = zlib.compress(dump.encode('utf-8'))
        self.json_file.write(dump_c)

    def add_archive(self, p_archive):
        """ Sets deep copy of p_archive as archive attribute. """
        self.archive = deepcopy(p_archive)

    def add_todolist(self, p_todolist):
        """ Sets deep copy of p_todolist as todolist attribute. """
        self.todolist = deepcopy(p_todolist)

    def save(self, p_todolist):
        """
        Saves a tuple with archive, todolist and command with its arguments
        into the backup file with unix timestamp as the key. Tuple is then
        indexed in backup file with combination of hash calculated from
        p_todolist and unix timestamp. Backup file is closed afterwards.
        """
        self._trim()

        current_hash = hash_todolist(p_todolist)
        list_todo = (self.todolist.print_todos()+'\n').splitlines(True)
        try:
            list_archive = (self.archive.print_todos()+'\n').splitlines(True)
        except AttributeError:
            list_archive = []

        self.backup_dict[self.timestamp] = (list_todo, list_archive,  self.label)

        index = self._get_index()
        index.insert(0, (self.timestamp, current_hash))
        self._save_index(index)

        self._write()
        self.close()

    def delete(self, p_timestamp=None, p_write=True):
        """ Removes backup from the backup file. """
        timestamp = p_timestamp or self.timestamp
        index = self._get_index()

        try:
            del self.backup_dict[timestamp]
            index.remove(index[[change[0] for change in index].index(timestamp)])
            self._save_index(index)

            if p_write:
                self._write()
        except KeyError:
            pass

    def _get_index(self):
        try:
            index = self.backup_dict['index']
        except KeyError:
            self.backup_dict['index'] = []
            index = self.backup_dict['index']

        return index

    def _save_index(self, p_index):
        """
        Saves index of backups supplied in p_index into the backup_file
        property with 'index' as the key.
        """
        self.backup_dict['index'] = p_index

    def _trim(self):
        """
        Removes oldest backups that exceed the limit configured in backup_count
        option.

        Does not write back to file system, make sure to call self._write()
        afterwards.
        """
        index = self._get_index()
        backup_limit = config().backup_count() - 1

        for changeset in index[backup_limit:]:
            self.delete(changeset[0], p_write=False)

    def read_backup(self, p_todolist=None, p_timestamp=None):
        """
        Retrieves a backup for p_timestamp or p_todolist (if p_timestamp is not
        specified) from backup file and sets timestamp, todolist, archive and
        label attributes to appropriate data from it.
        """
        if not p_timestamp:
            change_hash = hash_todolist(p_todolist)
            index = self._get_index()
            self.timestamp = index[[change[1] for change in index].index(change_hash)][0]
        else:
            self.timestamp = p_timestamp

        d = self.backup_dict[self.timestamp]

        self.todolist = TodoList(d[0])
        self.archive = TodoList(d[1])
        self.label = d[2]

    def apply(self, p_todolist, p_archive):
        """ Applies backup on supplied p_todolist. """
        if self.todolist and p_todolist:
            p_todolist.replace(self.todolist.todos())

        if self.archive and p_archive:
            p_archive.replace(self.archive.todos())

    def close(self):
        """ Closes backup file. """
        self.json_file.close()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2016 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from topydo.lib.Color import Color
from topydo.lib.Config import config
from topydo.lib.Recurrence import relative_date_to_date

# when a todo item has not enough information to determine the length, assume
# this length
ASSUMED_TODO_LENGTH = 14  # days

def progress_color(p_todo):
    color16_range = [
        2,   # green
        10,  # light green
        3,   # yellow
        1,   # red
    ]

    # https://commons.wikimedia.org/wiki/File:Xterm_256color_chart.svg
    # a gradient from green to yellow to red
    color256_range = \
        [22, 28, 34, 40, 46, 82, 118, 154, 190, 226, 220, 214, 208, 202, 196]

    def get_length():
        """
        Returns the length of the p_todo item in days, based on the recurrence
        period + due date, or the start/due date.
        """
        result = 0

        def diff_days(p_start, p_end):
            """
            Returns the difference in days between p_start and p_end, where
            start is before due.
            """
            diff = p_end - p_start
            return diff.days

        does_recur = p_todo.has_tag('rec')
        start_date = p_todo.start_date()
        due_date = p_todo.due_date()
        creation_date = p_todo.creation_date()

        if does_recur and due_date and not start_date:
            # add negation, offset is based on due date
            recurrence_pattern = p_todo.tag_value('rec')
            neg_recurrence_pattern = re.sub(r'^\+?', '-', recurrence_pattern)

            start = relative_date_to_date(neg_recurrence_pattern, due_date)
            result = diff_days(start, due_date)
        elif due_date and not start_date and not creation_date:
            result = ASSUMED_TODO_LENGTH
        elif due_date and start_date and due_date < start_date:
            result = ASSUMED_TODO_LENGTH
        elif due_date and not start_date and creation_date and due_date < creation_date:
            result = ASSUMED_TODO_LENGTH
        else:
            result = p_todo.length()

        # a todo item is at least one day long
        return max(1, result)

    def get_progress(p_todo, p_consider_parents=True):
        """
        Returns a value from 0 to 1 where we are today in a date range. Returns
        a value >1 when a todo item is overdue.
        """
        def progress_of_parents():
            try:
                parents = p_todo.parents()
            except AttributeError:
                parents = []

            if parents:
                return max(get_progress(parent, False) for parent in parents)
            else:
                return 0

        if p_todo.is_completed():
            return 0
        elif p_todo.is_overdue():
            return 1.1
        elif p_todo.due_date():
            days_till_due = p_todo.days_till_due()
            length = get_length()
            return max((length - days_till_due), 0) / length
        elif p_consider_parents:
            return progress_of_parents()
        else:
            return 0

    use_256_colors = config().colors() == 256
    color_range = color256_range if use_256_colors else color16_range
    progress = get_progress(p_todo)

    # TODO: remove linear scale to exponential scale
    if progress > 1:
        # overdue, return the last color
        return Color(color_range[-1])
    elif p_todo.is_completed():
        # return grey
        return Color(243) if use_256_colors else Color(7)
    else:
        # not overdue, calculate position over color range excl. due date
        # color
        pos = round(progress * (len(color_range) - 2))
        return Color(color_range[pos])



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides a completer class that can be used by the Column UI
CommmandLineWidget.
"""

from topydo.lib.Config import config
from topydo.ui.CompleterBase import CompleterBase, date_suggestions


class ColumnCompleter(CompleterBase):
    """
    Completer class that completes projects, contexts, dates and
    subcommands designed to work with CommandLineWidget for column UI.
    """
    def get_completions(self, p_word, p_is_first_word=False):
        def dates(p_word, p_tag):
            dates = []
            for date in date_suggestions():
                candidate = p_tag + ':' + date
                if candidate.startswith(p_word):
                    dates.append(candidate)

            return dates

        due_tag = config().tag_due()
        start_tag = config().tag_start()

        if p_word.startswith(due_tag + ':'):
            return dates(p_word, due_tag)
        elif p_word.startswith(start_tag + ':'):
            return dates(p_word, start_tag)
        else:
            return super().get_completions(p_word, p_is_first_word)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
A list of todo items.
"""

import math
import re
from datetime import date

from topydo.lib import Filter
from topydo.lib.Config import config
from topydo.lib.HashListValues import hash_list_values, max_id_length
from topydo.lib.printers.PrettyPrinter import PrettyPrinter
from topydo.lib.Todo import Todo
from topydo.lib.View import View


class InvalidTodoException(Exception):
    pass


class TodoListBase(object):
    """
    Provides operations for a todo list, such as adding items, removing them,
    etc.

    The list is usually a complete list found in the program's input (e.g. a
    todo.txt file), not an arbitrary set of todo items.
    """

    def __init__(self, p_todostrings):
        """
        Should be given a list of strings, each element a single todo string.
        The string will be parsed.
        """
        self._todos = []
        self._todo_id_map = {}
        self._id_todo_map = {}

        self.add_list(p_todostrings)
        self._dirty = False

    def __iter__(self):
        """
        Allows use of `for my_todo in todolist` constructs.
        """
        return iter(self._todos)

    def todo(self, p_identifier):
        """
        The _todos list has the same order as in the backend store (usually
        a todo.txt file. The user refers to the first task as number 1, so use
        index 0, etc.

        Alternative ways to identify a todo is using a hashed version based on
        the todo's text, or a regexp that matches the todo's source. The regexp
        match is a fallback.

        Returns None when the todo couldn't be found.
        """
        result = None

        def todo_by_uid(p_identifier):
            """ Returns the todo that corresponds to the unique ID. """
            result = None

            if config().identifiers() == 'text':
                try:
                    result = self._id_todo_map[p_identifier]
                except KeyError:
                    pass  # we'll try something else

            return result

        def todo_by_linenumber(p_identifier):
            """
            Attempts to find the todo on the given line number.

            When the identifier is a number but has leading zeros, the result
            will be None.
            """
            result = None

            if config().identifiers() != 'text':
                try:
                    if re.match(r'[1-9]\d*', p_identifier):
                        # the expression is a string and no leading zeroes,
                        # treat it as an integer
                        raise TypeError
                except TypeError as te:
                    try:
                        result = self._todos[int(p_identifier) - 1]
                    except (ValueError, IndexError):
                        raise InvalidTodoException from te

            return result

        def todo_by_regexp(p_identifier):
            """
            Returns the todo that is (uniquely) identified by the given regexp.
            If the regexp matches more than one item, no result is returned.
            """
            result = None

            candidates = Filter.GrepFilter(p_identifier).filter(self._todos)

            if len(candidates) == 1:
                result = candidates[0]
            else:
                raise InvalidTodoException

            return result

        result = todo_by_uid(p_identifier)

        if not result:
            result = todo_by_linenumber(p_identifier)

        if not result:
            # convert integer to text so we pass on a valid regex
            result = todo_by_regexp(str(p_identifier))

        return result

    def add(self, p_src):
        """
        Given a todo string, parse it and put it to the end of the list.
        """
        todos = self.add_list([p_src])

        return todos[0] if len(todos) else None

    def add_list(self, p_srcs):
        todos = [Todo(src) for src in p_srcs]
        if config().auto_delete_whitespace():
            todos = [todo for todo in todos if re.search(r'\S', todo.source())]
        self.add_todos(todos)

        return todos

    def add_todo(self, p_todo):
        """ Add an Todo object to the list. """
        self.add_todos([p_todo])

    def add_todos(self, p_todos):
        for todo in p_todos:
            self._todos.append(todo)

        self._update_todo_ids()
        self.dirty = True

    def delete(self, p_todo):
        """ Deletes a todo item from the list. """
        try:
            number = self._todos.index(p_todo)
            del self._todos[number]
            self._update_todo_ids()
            self.dirty = True
        except ValueError:
            # todo item couldn't be found, ignore
            pass

    def modify_todo(self, p_todo, p_new_source):
        """ Modify source of a Todo item from the list. """
        assert p_todo in self._todos
        p_todo.set_source_text(p_new_source)
        self._update_todo_ids()
        self.dirty = True

    def erase(self):
        """ Erases all todos from the list. """
        self._todos = []
        self.dirty = True

    def replace(self, p_todos):
        """ Replaces whole todolist with todo objects supplied as p_todos. """
        self.erase()
        self.add_todos(p_todos)
        self.dirty = True

    def count(self):
        """ Returns the number of todos on this list. """
        return len(self._todos)

    def append(self, p_todo, p_string):
        """
        Appends a text to the todo, specified by its number.
        The todo will be parsed again, such that tags and projects in de
        appended string are processed.
        """
        if len(p_string) > 0:
            new_text = p_todo.source() + ' ' + p_string
            p_todo.set_source_text(new_text)
            self._update_todo_ids()
            self.dirty = True

    def projects(self):
        """ Returns a set of all projects in this list. """
        result = set()
        for todo in self._todos:
            projects = todo.projects()
            result = result.union(projects)

        return result

    def contexts(self):
        """ Returns a set of all contexts in this list. """
        result = set()
        for todo in self._todos:
            contexts = todo.contexts()
            result = result.union(contexts)

        return result

    def view(self, p_sorter, p_filters):
        """
        Constructs a view of the todo list.

        A view is a sorted and filtered todo list, where the properties are
        defined by the end user. Todos is this list should not be modified,
        modifications should occur through this class.
        """
        return View(p_sorter, p_filters, self)

    @property
    def dirty(self):
        return self._dirty

    @dirty.setter
    def dirty(self, p_flag):
        self._dirty = p_flag

    def todos(self):
        return self._todos

    def set_todo_completed(self, p_todo, p_completion_date=date.today()):
        p_todo.set_completed(p_completion_date)
        self.dirty = True

    def set_priority(self, p_todo, p_priority):
        if p_todo.priority() != p_priority:
            p_todo.set_priority(p_priority)
            self.dirty = True

    def linenumber(self, p_todo):
        """
        Returns the line number of the todo item.
        """
        try:
            return self._todos.index(p_todo) + 1
        except ValueError as ex:
            raise InvalidTodoException from ex

    def uid(self, p_todo):
        """
        Returns the unique text-based ID for a todo item.
        """
        try:
            return self._todo_id_map[p_todo]
        except KeyError as ex:
            raise InvalidTodoException from ex

    def number(self, p_todo):
        """
        Returns the line number or text ID of a todo (depends on the
        configuration.
        """
        if config().identifiers() == "text":
            return self.uid(p_todo)
        else:
            return self.linenumber(p_todo)

    def max_id_length(self):
        """
        Returns the maximum length of a todo ID, used for formatting purposes.
        """
        if config().identifiers() == "text":
            return max_id_length(len(self._todos))
        else:
            try:
                return math.ceil(math.log(len(self._todos), 10))
            except ValueError:
                return 0


    def _update_todo_ids(self):
        # the idea is to have a hash that is independent of the position of the
        # todo. Use the text (without tags) of the todo to keep the id as
        # stable as possible (not influenced by priorities or due dates, etc.)
        self._todo_id_map = {}
        self._id_todo_map = {}

        uids = hash_list_values(self._todos, lambda t: t.text())

        for (todo, uid) in uids:
            self._todo_id_map[todo] = uid
            self._id_todo_map[uid] = todo

    def print_todos(self):
        """
        Returns a pretty-printed string (without colors) of the todo items in
        this list.
        """
        printer = PrettyPrinter()
        return "\n".join([str(s) for s in printer.print_list(self._todos)])

    def ids(self):
        """ Returns set with all todo IDs. """
        if config().identifiers() == 'text':
            ids = self._id_todo_map.keys()
        else:
            ids = [str(i + 1) for i in range(self.count())]
        return set(ids)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Entry file for the Python todo.txt CLI. """

import getopt
import sys

from topydo.ui.cli.CLI import CLIApplication
from topydo.ui.CLIApplicationBase import MAIN_LONG_OPTS, MAIN_OPTS, error

_WINDOWS = "win32" in sys.platform

# enable color on windows CMD
if _WINDOWS:
    import colorama
    colorama.init()


def main():
    """ Main entry point of the CLI. """
    try:
        args = sys.argv[1:]

        try:
            _, args = getopt.getopt(args, MAIN_OPTS, MAIN_LONG_OPTS)
        except getopt.GetoptError as e:
            error(str(e))
            sys.exit(1)

        if args[0] == 'prompt':
            try:
                from topydo.ui.prompt.Prompt import PromptApplication
                PromptApplication().run()
            except ImportError:
                error("Some additional dependencies for prompt mode were not installed, please install with 'pip3 install topydo[prompt]'")
        elif args[0] == 'columns':
            try:
                from topydo.ui.columns.Main import UIApplication
                UIApplication().run()
            except ImportError:
                error("Some additional dependencies for column mode were not installed, please install with 'pip3 install topydo[columns]'")
            except NameError as err:
                if _WINDOWS:
                    error("Column mode is not supported on Windows.")
                else:
                    error("Could not load column mode: {}".format(err))
        else:
            CLIApplication().run()
    except IndexError:
        CLIApplication().run()

if __name__ == '__main__':
    main()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import os
import re
import shlex
from collections import OrderedDict
from functools import lru_cache
from itertools import accumulate
from string import ascii_lowercase

from topydo.lib.Color import Color


def home_config_path(p_filename):
    return os.path.join(os.path.expanduser('~'), p_filename)

class ConfigError(Exception):
    def __init__(self, p_text):
        self.text = p_text

    def __str__(self):
        return self.text


class _Config:
    def __init__(self, p_path=None, p_overrides=None):
        """
        Constructor.

        If p_path is given, that is the only configuration file that will be
        read.

        If p_overrides is given, some options are ultimately overridden. This
        is for some command line options which override any configuration file
        (such as todo.txt location passed with -t). The key is a tuple of
        (section, option), the value is the option's value.
        """
        self.sections = [
            'add',
            'aliases',
            'colorscheme',
            'column_keymap',
            'columns',
            'dep',
            'ls',
            'sort',
            'tags',
            'topydo',
        ]

        self.defaults = {
            'topydo': {
                'default_command': 'ls',
                'colors': 'auto',
                'force_colors': '0',
                'filename': 'todo.txt',
                'archive_filename': 'done.txt',
                'identifiers': 'linenumber',
                'identifier_alphabet': '0123456789abcdefghijklmnopqrstuvwxyz',
                'backup_count': '5',
                'auto_delete_whitespace': '1',
            },

            'add': {
                'auto_creation_date': '1',
            },

            'ls': {
                'hide_tags': 'id,p,ical',
                'hidden_item_tags': 'h,hide',
                'indent': '0',
                'list_limit': '-1',
                'list_format': '|%I| %x %{(}p{)} %c %s %k %{due:}d %{t:}t',
            },

            'tags': {
                'tag_start': 't',
                'tag_due': 'due',
                'tag_star': 'star',
            },

            'sort': {
                'keep_sorted': '0',
                'sort_string': 'desc:completed,desc:importance,due,desc:priority',
                'group_string': '',
                'ignore_weekends': '1',
            },

            'dep': {
                'append_parent_projects': '0',
                'append_parent_contexts': '0',
            },

            'edit': {
            },

            'colorscheme': {
                'project_color': 'red',
                'context_color': 'magenta',
                'metadata_color': 'green',
                'link_color': 'cyan',
                'priority_colors': 'A:cyan,B:yellow,C:blue',
                'focus_background_color': 'gray',
                'marked_background_color': 'blue'
            },

            'aliases': {
                'lsproj': 'lsprj',
                'listprj': 'lsprj',
                'listproj': 'lsprj',
                'listproject': 'lsprj',
                'listprojects': 'lsprj',
                'listcon': 'lscon',
                'listcontext': 'lscon',
                'listcontexts': 'lscon',
            },

            'columns': {
                'column_width': '40',
            },

            'column_keymap': {
                'gg': 'home',
                'G': 'end',
                'j': 'down',
                'k': 'up',
                'd': 'cmd del {}',
                'e': 'cmd edit {}',
                'u': 'cmd revert',
                'x': 'cmd do {}',
                'm': 'mark',
                '<C-a>': 'mark_all',
                '.': 'repeat',
                'pp': 'postpone',
                'ps': 'postpone_s',
                'pr': 'pri',
                '0': 'first_column',
                '$': 'last_column',
                'h': 'prev_column',
                'l': 'next_column',
                'A': 'append_column',
                'I': 'insert_column',
                'E': 'edit_column',
                'D': 'delete_column',
                'Y': 'copy_column',
                'L': 'swap_left',
                'R': 'swap_right',
                '<Left>': 'prev_column',
                '<Right>': 'next_column',
                '<Down>': 'down',
                '<Esc>': 'reset',
            },
        }

        self.config = {}

        self.cp = configparser.RawConfigParser(strict=False)
        # allow uppercase config keys
        self.cp.optionxform = lambda option: option

        for section in self.defaults:
            self.cp.add_section(section)

            for option, value in self.defaults[section].items():
                self.cp.set(section, option, value)

        files = [
            "/etc/topydo.conf",
            home_config_path('.config/topydo/config'),
            home_config_path('.topydo'),
            ".topydo",
            "topydo.conf",
            "topydo.ini",
        ]

        # when a path is given, *only* use the values in that file, or the
        # defaults listed above.
        if p_path is not None:
            files = [p_path]

        self.cp.read(files)
        self._supplement_sections()

        if p_overrides:
            for (section, option), value in p_overrides.items():
                self.cp.set(section, option, value)

    def _supplement_sections(self):
        for section in self.sections:
            if not self.cp.has_section(section):
                self.cp.add_section(section)

    def default_command(self):
        return self.cp.get('topydo', 'default_command')

    def colors(self, p_hint_possible=True):
        """
        Returns 0, 16 or 256 representing the number of colors that should be
        used in the output.

        A hint can be passed whether the device that will output the text
        supports colors.
        """
        lookup = {
            'false': 0,
            'no': 0,
            '0': 0,
            '1': 16,
            'true': 16,
            'yes': 16,
            '16': 16,
            '256': 256,
        }

        try:
            forced = self.cp.get('topydo', 'force_colors') == '1'
        except ValueError:
            forced = self.defaults['topydo']['force_colors'] == '1'

        try:
            colors = lookup[self.cp.get('topydo', 'colors').lower()]  # pylint: disable=no-member
        except ValueError:
            colors = lookup[self.defaults['topydo']['colors'].lower()]  # pylint: disable=no-member
        except KeyError:
            # for invalid values or 'auto'
            colors = 16 if p_hint_possible else 0

        # disable colors when no colors are enforced on the commandline and
        # color support is determined automatically
        return 0 if not forced and not p_hint_possible else colors

    def todotxt(self):
        return os.path.expanduser(self.cp.get('topydo', 'filename'))

    def archive(self):
        return os.path.expanduser(self.cp.get('topydo', 'archive_filename'))

    def identifiers(self):
        return self.cp.get('topydo', 'identifiers')

    def backup_count(self):
        try:
            value = self.cp.getint('topydo', 'backup_count')
            if value < 0:
                value = 0
            return value
        except ValueError:
            return int(self.defaults['topydo']['backup_count'])

    def auto_delete_whitespace(self):
        try:
            return self.cp.getboolean('topydo', 'auto_delete_whitespace')
        except ValueError:
            return self.defaults['topydo']['auto_delete_whitespace'] == '1'

    def list_limit(self):
        try:
            return self.cp.getint('ls', 'list_limit')
        except ValueError:
            return int(self.defaults['ls']['list_limit'])

    def list_indent(self):
        try:
            return self.cp.getint('ls', 'indent')
        except ValueError:
            return int(self.defaults['ls']['indent'])

    def keep_sorted(self):
        try:
            return self.cp.getboolean('sort', 'keep_sorted')
        except ValueError:
            return self.defaults['sort']['keep_sorted'] == '1'

    def sort_string(self):
        return self.cp.get('sort', 'sort_string')

    def group_string(self):
        return self.cp.get('sort', 'group_string')

    def ignore_weekends(self):
        try:
            return self.cp.getboolean('sort', 'ignore_weekends')
        except ValueError:
            return self.defaults['sort']['ignore_weekends'] == '1'

    def append_parent_projects(self):
        try:
            return self.cp.getboolean('dep', 'append_parent_projects')
        except ValueError:
            return self.defaults['dep']['append_parent_projects'] == '1'

    def append_parent_contexts(self):
        try:
            return self.cp.getboolean('dep', 'append_parent_contexts')
        except ValueError:
            return self.defaults['dep']['append_parent_contexts'] == '1'

    def _get_tag(self, p_tag):
        try:
            return self.config[p_tag]
        except KeyError:
            value = self.cp.get('tags', p_tag)
            self.config[p_tag] = value
            return value

    def tag_due(self):
        return self._get_tag('tag_due')

    def tag_start(self):
        return self._get_tag('tag_start')

    def tag_star(self):
        return self._get_tag('tag_star')

    def hidden_tags(self):
        """ Returns a list of tags to be hidden from the 'ls' output. """
        hidden_tags = self.cp.get('ls', 'hide_tags')
        # pylint: disable=no-member
        return [] if hidden_tags == '' else [tag.strip() for tag in
                                             hidden_tags.split(',')]

    def hidden_item_tags(self):
        """ Returns a list of tags which hide an item from the 'ls' output. """
        hidden_item_tags = self.cp.get('ls', 'hidden_item_tags')
        # pylint: disable=no-member
        return [] if hidden_item_tags == '' else [tag.strip() for tag in
                                                  hidden_item_tags.split(',')]

    @lru_cache(maxsize=26)
    def priority_color(self, p_priority):
        """
        Returns a dict with priorities as keys and color numbers as value.
        """
        def _str_to_dict(p_string):
            pri_colors_dict = dict()
            for pri_color in p_string.split(','):
                pri, color = pri_color.split(':')
                pri_colors_dict[pri] = Color(color)

            return pri_colors_dict

        try:
            pri_colors_str = self.cp.get('colorscheme', 'priority_colors')

            if pri_colors_str == '':
                pri_colors_dict = _str_to_dict('A:-1,B:-1,C:-1')
            else:
                pri_colors_dict = _str_to_dict(pri_colors_str)
        except ValueError:
            pri_colors_dict = _str_to_dict(self.defaults['colorscheme']['priority_colors'])

        return pri_colors_dict[p_priority] if p_priority in pri_colors_dict else Color('NEUTRAL')

    def project_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'project_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'project_color'))

    def context_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'context_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'context_color'))

    def metadata_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'metadata_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'metadata_color'))

    def link_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'link_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'link_color'))

    def focus_background_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'focus_background_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'focus_background_color'))

    def marked_background_color(self):
        try:
            return Color(self.cp.getint('colorscheme', 'marked_background_color'))
        except ValueError:
            return Color(self.cp.get('colorscheme', 'marked_background_color'))

    def auto_creation_date(self):
        try:
            return self.cp.getboolean('add', 'auto_creation_date')
        except ValueError:
            return self.defaults['add']['auto_creation_date'] == '1'

    @lru_cache(maxsize=1)
    def aliases(self):
        """
        Returns dict with aliases names as keys and pairs of actual
        subcommand and alias args as values.
        """
        aliases = self.cp.items('aliases')
        alias_dict = dict()

        for alias, meaning in aliases:
            try:
                meaning = shlex.split(meaning)
                real_subcommand = meaning[0]
                alias_args = meaning[1:]
                alias_dict[alias] = (real_subcommand, alias_args)
            except ValueError as verr:
                alias_dict[alias] = str(verr)

        return alias_dict

    def list_format(self):
        """ Returns the list format used by `ls` """
        return self.cp.get('ls', 'list_format')

    def column_width(self):
        try:
            width = self.cp.getint('columns', 'column_width')

            if width < 1:
                # read default
                raise ValueError

            return width
        except ValueError:
            return int(self.defaults['columns']['column_width'])

    @lru_cache(maxsize=1)
    def column_keymap(self):
        """ Returns keymap and keystates used in column mode """
        keystates = set()

        shortcuts = self.cp.items('column_keymap')
        keymap_dict = dict(shortcuts)

        for combo, action in shortcuts:
            # add all possible prefixes to keystates
            combo_as_list = re.split('(<[A-Z].+?>|.)', combo)[1::2]
            if len(combo_as_list) > 1:
                keystates |= set(accumulate(combo_as_list[:-1]))

            if action in ['pri', 'postpone', 'postpone_s']:
                keystates.add(combo)

            if action == 'pri':
                for c in ascii_lowercase:
                    keymap_dict[combo + c] = 'cmd pri {} ' + c

        return (keymap_dict, keystates)

    def editor(self):
        """
        Returns the editor to invoke. It returns a list with the command in
        the first position and its arguments in the remainder.
        """
        result = 'vi'
        if 'TOPYDO_EDITOR' in os.environ and os.environ['TOPYDO_EDITOR']:
            result = os.environ['TOPYDO_EDITOR']
        else:
            try:
                result = str(self.cp.get('edit', 'editor'))
            except configparser.NoOptionError:
                if 'EDITOR' in os.environ and os.environ['EDITOR']:
                    result = os.environ['EDITOR']

        return shlex.split(result)

    def identifier_alphabet(self):
        alphabet = self.cp.get('topydo', 'identifier_alphabet')

        # deduplicate characters alphabet. Use a dictionary, but an ordered one
        # to keep determinism.
        return list(OrderedDict([(c, None) for c in alphabet]).keys())

def config(p_path=None, p_overrides=None):
    """
    Retrieve the config instance.

    If a path is given, the instance is overwritten by the one that supplies an
    additional filename (for testability). Moreover, no other configuration
    files will be read when a path is given.

    Overrides will discard a setting in any configuration file and use the
    passed value instead. Structure: (section, option) => value
    The previous configuration instance will be discarded.
    """
    if not config.instance or p_path is not None or p_overrides is not None:
        try:
            config.instance = _Config(p_path, p_overrides)
        except configparser.ParsingError as perr:
            raise ConfigError(str(perr)) from perr

    return config.instance

config.instance = None



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import arrow

from topydo.lib import TodoFile, TodoList
from topydo.lib.ChangeSet import ChangeSet
from topydo.lib.Command import Command, InvalidCommandArgument
from topydo.lib.Config import config


class RevertCommand(Command):
    def __init__(self, p_args, p_todolist,  # pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(p_args, p_todolist, p_out, p_err, p_prompt)

        self._backup = None
        self._archive_file = None
        self._archive = None

    def execute(self):
        if not super().execute():
            return False

        self._backup = ChangeSet()
        archive_path = config().archive()
        if archive_path:
            self._archive_file = TodoFile.TodoFile(config().archive())
            self._archive = TodoList.TodoList(self._archive_file.read())

        if len(self.args) > 1:
            self.error(self.usage())
        else:
            try:
                arg = self.argument(0)
                self._handle_args(arg)
            except InvalidCommandArgument:
                try:
                    self._revert_last()
                except (ValueError, KeyError):
                    self.error('No backup was found for the current state of '
                               + config().todotxt())

        self._backup.close()

    def _revert(self, p_timestamp=None):
        self._backup.read_backup(self.todolist, p_timestamp)
        self._backup.apply(self.todolist, self._archive)

        if self._archive:
            self._archive_file.write(self._archive.print_todos())

        self.out("Reverted to state before: " + self._backup.label)

    def _revert_last(self):
        self._revert()
        self._backup.delete()

    def _revert_to_specific(self, p_position):
        timestamps = [timestamp for timestamp, _ in self._backup]
        position = int(p_position) - 1  # numbering in UI starts with 1
        try:
            timestamp = timestamps[position]
            self._revert(timestamp)
            for timestamp in timestamps[:position + 1]:
                self._backup.read_backup(p_timestamp=timestamp)
                self._backup.delete()
        except IndexError:
            self.error('Specified index is out range')

    def _handle_args(self, p_arg):
        try:
            if p_arg == 'ls':
                self._handle_ls()
            elif p_arg.isdigit():
                self._revert_to_specific(p_arg)
            else:
                raise InvalidCommandArgument
        except InvalidCommandArgument:
            self.error(self.usage())

    def _handle_ls(self):
        num = 1
        for timestamp, change in self._backup:
            label = change[2]
            time = arrow.get(float(timestamp)).format('YYYY-MM-DD HH:mm:ss')

            self.out('{0: >3}| {1} | {2}'.format(str(num), time, label))
            num += 1

    def usage(self):
        return """Synopsis:
  revert [ls]
  revert [NUMBER]"""

    def help(self):
        return """\
Reverts last commands.

* ls       : Lists all backups ordered and numbered chronologically (starting
             with 1 for the latest backup).
* [NUMBER] : revert to specific point in history specified by NUMBER.

Output example for `revert ls`:
1 | 1970-01-01 00:00:02 | add Baz
2 | 1970-01-01 00:00:01 | add Bar
3 | 1970-01-01 00:00:00 | add Foo

In such example executing `revert 2` will revert the todo and archive files to
the state before execution of `add Bar`.

* `revert` without any further arguments will revert to the latest backup
  available, provided that this backup matches the current state of the todo
  file.
  Topydo will refuse to revert, if any changes to todo file were made by
  external application after the latest backup. To force a `revert` action use
  it with a NUMBER.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the class that represents a single todo item.
"""

import re
from datetime import date

from topydo.lib.TodoParser import parse_line
from topydo.lib.Utils import is_valid_priority


class TodoBase(object):
    """
    This class represents a single todo item in a todo.txt file. It maintains
    an internal data dictionary of various attributes, but also keeps the plain
    text format in shape such that it can be printed back to a file with as few
    distortions as possible (no re-shuffling of attributes).

    This is a base class, but supports enough to process any item in a todo.txt
    file. Derived classes add some interpretation to the tags that may appear
    in a todo item.
    """

    def __init__(self, p_src):
        self.src = ""
        self.fields = {}

        self.set_source_text(p_src)

    def tag_value(self, p_key, p_default=None):
        """
        Returns a tag value associated with p_key. Returns p_default if p_key
        does not exist (which defaults to None).
        """
        return self.tag_values(p_key)[0] if p_key in self.fields['tags'] else p_default

    def tag_values(self, p_key):
        """
        Returns a list of all tag values associated with p_key. Returns
        empty list if p_key does not exist.
        """
        return self.fields['tags'][p_key] if p_key in self.fields['tags'] else []

    def has_tag(self, p_key, p_value=""):
        """
        Returns true when there is at least one tag with the given key. If a
        value is passed, it will only return true when there exists a tag with
        the given key-value combination.
        """
        tags = self.fields['tags']
        return p_key in tags and (p_value == "" or p_value in tags[p_key])

    def add_tag(self, p_key, p_value):
        """ Adds a tag to the todo. """
        self.set_tag(p_key, p_value, True)

    def _remove_tag_helper(self, p_key, p_value):
        """
        Removes a tag from the internal todo dictionary. Only those instances
        with the given value are removed. If the value is empty, all tags with
        the given key are removed.
        """
        tags = self.fields['tags']

        try:
            tags[p_key] = [t for t in tags[p_key] if p_value != "" and t != p_value]
            if len(tags[p_key]) == 0:
                del tags[p_key]
        except KeyError:
            pass

    def set_tag(self, p_key, p_value="", p_force_add=False, p_old_value=""):
        """
        Sets a occurrence of the tag identified by p_key. Sets an arbitrary
        instance of the tag when the todo contains multiple tags with this key.
        When p_key does not exist, the tag is added.

        When p_value is not set, the tag will be removed.

        When p_force_add is true, a tag will always be added to the todo, in
        case there is already a tag with the given key.

        When p_old_value is set, all tags having this value will be set to the
        new value.
        """
        if p_value == "":
            self.remove_tag(p_key, p_old_value)
            return

        tags = self.fields['tags']
        value = p_old_value if p_old_value else self.tag_value(p_key)

        if not p_force_add and value:
            self._remove_tag_helper(p_key, value)

            self.src = re.sub(
                r'\b' + p_key + ':' + value + r'\b',
                p_key + ':' + p_value,
                self.src
            )
        else:
            self.src += ' ' + p_key + ':' + p_value

        try:
            tags[p_key].append(p_value)
        except KeyError:
            tags[p_key] = [p_value]

    def remove_tag(self, p_key, p_value=""):
        """
        Removes a tag from the todo.
        When the value is empty (default), all occurrences of the tag will be
        removed.
        Else, only those tags with the value will be removed.
        """
        self._remove_tag_helper(p_key, p_value)

        # when value == "", match any value having key p_key
        value = p_value if p_value != "" else r'\S+'
        self.src = re.sub(r'\s?\b' + p_key + ':' + value + r'\b', '', self.src)

    def tags(self):
        """
        Returns a list of tuples with key-value pairs representing tags in
        this todo item.
        """
        tags = self.fields['tags']
        return [(t, v) for t in tags for v in tags[t]]

    def set_priority(self, p_priority):
        """
        Sets the priority of the todo. Must be a single capital letter [A-Z],
        or None to unset the priority.
        Priority remains unchanged when an invalid priority is given, or when
        the task was completed.
        """
        if not self.is_completed() and (p_priority is None or
                                        is_valid_priority(p_priority)):
            self.fields['priority'] = p_priority

            priority_str = '' if p_priority is None else '(' + p_priority + ') '
            self.src = re.sub(r'^(\([A-Z]\) )?', priority_str, self.src)

    def priority(self):
        """
        Returns the priority of this todo, or None if no priority is set.
        """
        return self.fields['priority']

    def text(self, p_with_tags=False):
        """ Returns the todo text with tags stripped off. """
        return self.src if p_with_tags else self.fields['text']

    def source(self):
        """
        Returns the source text of the todo. This is the raw text with all
        the tags included.
        """
        return self.text(True)

    def set_source_text(self, p_text):
        """ Sets the todo source text. The text will be parsed again. """
        self.src = p_text.strip()
        self.fields = parse_line(self.src)

    def projects(self):
        """ Returns a set of projects associated with this todo item. """
        return set(self.fields['projects'])

    def contexts(self):
        """ Returns a set of contexts associated with this todo item. """
        return set(self.fields['contexts'])

    def is_completed(self):
        """ Returns True iff this todo has been completed. """
        return self.fields['completed']

    def completion_date(self):
        """
        Returns the completion date when the todo has been completed, or None
        otherwise.
        """
        return self.fields['completionDate']

    def set_completed(self, p_completion_date=date.today()):
        """
        Marks the todo as complete.
        Sets the completed flag and sets the completion date to today.
        """
        if not self.is_completed():
            self.set_priority(None)

            self.fields['completed'] = True
            self.fields['completionDate'] = p_completion_date

            self.src = re.sub(r'^(\([A-Z]\) )?',
                              'x ' + p_completion_date.isoformat() + ' ',
                              self.src)

    def set_creation_date(self, p_date=date.today()):
        """
        Sets the creation date of a todo. Should be passed a date object.
        """
        self.fields['creationDate'] = p_date

        # not particularly pretty, but inspired by
        # http://bugs.python.org/issue1519638 non-existent matches trigger
        # exceptions, hence the lambda
        self.src = re.sub(
            r'^(x \d{4}-\d{2}-\d{2} |\([A-Z]\) )?(\d{4}-\d{2}-\d{2} )?(.*)$',
            lambda m:
            u"{}{} {}".format(m.group(1) or '', p_date.isoformat(),
                                m.group(3)), self.src)

    def creation_date(self):
        """ Returns the creation date of a todo. """
        return self.fields['creationDate']



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from itertools import chain

from topydo.lib.prettyprinters.Colors import PrettyPrinterColorFilter
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.TopydoString import TopydoString


class Printer(object):
    """
    An abstract class that turns todo items into strings.

    Subclasses must at least implement the print_todo method.
    """

    def print_todo(self, p_todo):
        raise NotImplementedError

    def print_list(self, p_todos):
        result = ''

        for todo in p_todos:
            result += self.print_todo(todo)

        return result

    def print_groups(self, p_groups):
        todos = list(chain.from_iterable(p_groups.values()))
        return self.print_list(todos)


class PrettyPrinter(Printer):
    """
    Prints todo items on a single line, decorated by the filters passed by
    the caller.

    The caller can adjust the output by passing on a set of filters, that may
    add colors, indentation, etc. These filters are found in the
    PrettyPrinterFilter module.
    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        self.filters = []

    def add_filter(self, p_filter):
        """
        Adds a filter to be applied when calling print_todo.

        p_filter is an instance of a PrettyPrinterFilter.
        """
        self.filters.append(p_filter)

    def print_todo(self, p_todo):
        """ Given a todo item, pretty print it. """
        todo_str = p_todo.source()

        for ppf in self.filters:
            todo_str = ppf.filter(todo_str, p_todo)

        return TopydoString(todo_str)

    def print_list(self, p_todos):
        """
        Given a list of todo items, pretty print it and return a list of
        formatted TopydoStrings. The output function in the UI should convert
        the colors inside properly.
        """
        return [self.print_todo(todo) for todo in p_todos]

    def print_groups(self, p_groups):
        result = []
        first = True

        def print_header(p_key):
            """ Prints a header for the given key. """
            if not first:
                result.append('')

            key_string = ", ".join(p_key)
            result.append(key_string)
            result.append("=" * len(key_string))

        for key, todos in p_groups.items():
            if key != ():
                # don't print a header for the case that no valid grouping
                # could be made (e.g. an invalid group expression)
                print_header(key)

            first = False
            result += self.print_list(todos)

        return [TopydoString(s) for s in result]

def pretty_printer_factory(p_todolist, p_additional_filters=None):
    """ Returns a pretty printer suitable for the ls and dep subcommands. """
    p_additional_filters = p_additional_filters or []

    printer = PrettyPrinter()
    printer.add_filter(PrettyPrinterNumbers(p_todolist))

    for ppf in p_additional_filters:
        printer.add_filter(ppf)

    # apply colors at the last step, the ANSI codes may confuse the
    # preceding filters.
    printer.add_filter(PrettyPrinterColorFilter())

    return printer



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date, timedelta

from topydo.lib.Config import config
from topydo.lib.MultiCommand import MultiCommand
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.Utils import date_string_to_date


class PostponeCommand(MultiCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.move_start_date = False
        self.last_argument = True

    def get_flags(self):
        return("s", [])

    def process_flag(self, p_option, p_value):
        if p_option == '-s':
            self.move_start_date = True

    def _execute_multi_specific(self):
        def _get_offset(p_todo):
            offset = p_todo.tag_value(
                config().tag_due(), date.today().isoformat())

            offset_date = date_string_to_date(offset)

            if offset_date < date.today():
                offset_date = date.today()

            return offset_date

        pattern = self.args[-1]
        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))

        for todo in self.todos:
            try:
                offset = _get_offset(todo)
            except ValueError:
                self.error("Postponing todo item failed: invalid due date.")
                break

            new_due = relative_date_to_date(pattern, offset)

            if new_due:
                if self.move_start_date and todo.start_date():
                    length = todo.length()
                    new_start = new_due - timedelta(length)
                    # pylint: disable=E1103
                    todo.set_tag(config().tag_start(), new_start.isoformat())
                elif self.move_start_date and not todo.start_date():
                    self.error("Warning: todo item has no (valid) start date, therefore it was not adjusted.")

                # pylint: disable=E1103
                todo.set_tag(config().tag_due(), new_due.isoformat())

                self.todolist.dirty = True
                self.out(self.printer.print_todo(todo))
            else:
                self.error("Invalid date pattern given.")
                break

    def usage(self):
        return """\
Synopsis: postpone [-s] <NUMBER> [<NUMBER2> ...] <PATTERN>
          postpone [-x] -e <EXPRESSION>\
"""

    def help(self):
        return """\
Postpone the todo item(s) with the given NUMBER(s) and the given PATTERN.

Postponing is done by adjusting the due date(s) of the todo(s), and if the -s
flag is given, the start date accordingly.

It is also possible to postpone items as complete with an EXPRESSION using
the -e flag. Use -x to also process todo items that are normally invisible (as
with the 'ls' subcommand).

The PATTERN is a relative date, written in the format <COUNT> <PERIOD> where
COUNT is a number and PERIOD is either 'd', 'w', 'm' or 'y', which stands for
days, weeks, months and years respectively. Example: 'postpone 1 2w' postpones
todo number 1 for 2 weeks.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from functools import lru_cache

from topydo.Commands import SUBCOMMAND_MAP
from topydo.lib.Config import config


@lru_cache(maxsize=1)
def _get_subcmds():
    subcmd_map = config().aliases().copy()
    subcmd_map.update(SUBCOMMAND_MAP)

    return sorted(subcmd_map.keys())


def date_suggestions():
    """
    Returns a list of relative date that is presented to the user as auto
    complete suggestions.
    """
    # don't use strftime, prevent locales to kick in
    days_of_week = {
        0: "Monday",
        1: "Tuesday",
        2: "Wednesday",
        3: "Thursday",
        4: "Friday",
        5: "Saturday",
        6: "Sunday"
    }

    dates = [
        'today',
        'tomorrow',
    ]

    # show days of week up to next week
    dow = datetime.date.today().weekday()
    for i in range(dow + 2 % 7, dow + 7):
        dates.append(days_of_week[i % 7])

    # and some more relative days starting from next week
    dates += ["1w", "2w", "1m", "2m", "3m", "1y"]

    return dates


class CompleterBase(object):
    def __init__(self, p_todolist):
        self.todolist = p_todolist
        self._all_subcmds = _get_subcmds()

    def _contexts(self, p_word):
        completions = ['@' + context for context in self.todolist.contexts() if
                       context.startswith(p_word[1:])]
        return sorted(completions)

    def _projects(self, p_word):
        completions = ['+' + project for project in self.todolist.projects() if
                       project.startswith(p_word[1:])]
        return sorted(completions)

    def _subcmds(self, p_word):
        completions = [cmd for cmd in self._all_subcmds if
                       cmd.startswith(p_word)]
        return completions

    def get_completions(self, p_word, p_is_first_word=False):
        completions = []

        if p_word.startswith('+'):
            completions = self._projects(p_word)
        elif p_word.startswith('@'):
            completions = self._contexts(p_word)
        elif p_is_first_word:
            completions = self._subcmds(p_word)

        return completions



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import InvalidCommandArgument
from topydo.lib.Config import config
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.TodoListBase import InvalidTodoException
from topydo.lib.TodoParser import parse_line
from topydo.lib.WriteCommand import WriteCommand


class AppendCommand(WriteCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(p_args, p_todolist, p_out, p_err,
                p_prompt)

    def execute(self):
        if not super().execute():
            return False

        try:
            number = self.argument(0)
            text = " ".join(self.args[1:])

            if text:
                todo = self.todolist.todo(number)
                new_text_parsed = parse_line(text)
                new_tags = new_text_parsed['tags']
                for tag in (config().tag_start(), config().tag_due()):
                    if tag in new_tags:
                        todo.remove_tag(tag)
                self.todolist.append(todo, text)
                self.postprocess_input_todo(todo)

                self.printer.add_filter(PrettyPrinterNumbers(self.todolist))
                self.out(self.printer.print_todo(todo))
            else:
                self.error(self.usage())
        except InvalidCommandArgument:
            self.error(self.usage())
        except InvalidTodoException:
            self.error("Invalid todo number given.")

    def usage(self):
        return """Synopsis: append <NUMBER> <TEXT>"""

    def help(self):
        return """\
Adds the given TEXT to the end of the todo indicated by NUMBER.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module that calculates identifiers for each item in a list, based on the hash
value of each item.
"""

from hashlib import sha1

from topydo.lib.Config import config

_DEFAULT_ALPHABET = '0123456789abcdefghijklmnopqrstuvwxyz'

# a two-dimensional lookup table, the first dimension is the length of the
# configured alphabet, the second dimension is the width of the ID
# The values are prime numbers that are used for populating the hash table.
_TABLE_SIZES = {
    10: {3: 997, 4: 9973, 5: 99991, 6: 999983},
    11: {3: 1327, 4: 14639, 5: 161047, 6: 1771559},
    12: {3: 1723, 4: 20731, 5: 248827, 6: 2985979},
    13: {3: 2179, 4: 28559, 5: 371291, 6: 4826797},
    14: {3: 2741, 4: 38393, 5: 573811, 6: 7529519},
    15: {3: 3373, 4: 50599, 5: 759371, 6: 11390593},
    16: {3: 4093, 4: 65521, 5: 1048573},
    17: {3: 4909, 4: 83497, 5: 1419839},
    18: {3: 5827, 4: 104971, 5: 1889561},
    19: {3: 6857, 4: 130307, 5: 2476081},
    20: {3: 7993, 4: 159979, 5: 3199997},
    21: {3: 9257, 4: 194479, 5: 4084081},
    22: {3: 10639, 4: 245239, 5: 5153623},
    23: {3: 12163, 4: 279823, 5: 6436327},
    24: {3: 13807, 4: 331769, 5: 7962607},
    25: {3: 15619, 4: 390581, 5: 9765619},
    26: {3: 17573, 4: 456959, 5: 11881357},
    27: {3: 19681, 4: 531383, 5: 14348891},
    28: {3: 21943, 4: 614639, 5: 17210353},
    29: {3: 24379, 4: 707279, 5: 20511143},
    30: {3: 26993, 4: 809993, 5: 24299981},
    31: {3: 29789, 4: 923513, 5: 28629149},
    32: {3: 32749, 4: 1048573},
    33: {3: 35933, 4: 1185907},
    34: {3: 39301, 4: 1336333},
    35: {3: 42863, 4: 1500619},
    36: {3: 46649, 4: 1679609},
    37: {3: 50651, 4: 1874153},
    38: {3: 54869, 4: 2085133},
    39: {3: 59281, 4: 2313439},
    40: {3: 63997, 4: 2559989},
    41: {3: 68917, 4: 2825759},
    42: {3: 74077, 4: 3111679},
    43: {3: 79493, 4: 3418799},
    44: {3: 85159, 4: 3748079},
    45: {3: 91121, 4: 4100611},
    46: {3: 97327, 4: 4477453},
    47: {3: 103813, 4: 4879669},
    48: {3: 110587, 4: 5308379},
    49: {3: 117643, 4: 5764799},
    50: {3: 124991, 4: 6249989},
}


class _TableSizeException(Exception):
    pass

def _get_table_size(p_alphabet, p_num):
    """
    Returns a prime number that is suitable for the hash table size. The size
    is dependent on the alphabet used, and the number of items that need to be
    hashed. The table size is at least 100 times larger than the number of
    items to be hashed, to avoid collisions.

    When the alphabet is too little or too large, then _TableSizeException is
    raised. Currently an alphabet of 10 to 40 characters is supported.
    """
    try:
        for width, size in sorted(_TABLE_SIZES[len(p_alphabet)].items()):
            if p_num < size * 0.01:
                return width, size
    except KeyError:
        pass

    raise _TableSizeException('Could not find appropriate table size for given alphabet')

def hash_list_values(p_list, p_key=lambda i: i):  # pragma: no branch
    """
    Calculates a unique value for each item in the list, these can be used as
    identifiers.

    The value is based on hashing an item using the p_key function.

    Suitable for lists not larger than approx. 16K items.

    Returns a tuple with the status and a list of tuples where each item is
    combined with the ID.
    """
    def to_base(p_alphabet, p_value):
        """
        Converts integer to text ID with characters from the given alphabet.

        Based on answer at
        https://stackoverflow.com/questions/1181919/python-base-36-encoding
        """
        result = ''
        while p_value:
            p_value, i = divmod(p_value, len(p_alphabet))
            result = p_alphabet[i] + result

        return result or p_alphabet[0]

    result = []
    used = set()
    alphabet = config().identifier_alphabet()

    try:
        _, size = _get_table_size(alphabet, len(p_list))
    except _TableSizeException:
        alphabet = _DEFAULT_ALPHABET
        _, size = _get_table_size(alphabet, len(p_list))

    for item in p_list:
        # obtain the to-be-hashed value
        raw_value = p_key(item)

        # hash
        hasher = sha1()
        hasher.update(raw_value.encode('utf-8'))
        hash_value = int(hasher.hexdigest(), 16) % size

        # resolve possible collisions
        while hash_value in used:
            hash_value = (hash_value + 1) % size

        used.add(hash_value)
        result.append((item, to_base(alphabet, hash_value)))

    return result

def max_id_length(p_num):
    """
    Returns the length of the IDs used, given the number of items that are
    assigned an ID. Used for padding in lists.
    """
    try:
        alphabet = config().identifier_alphabet()
        length, _ = _get_table_size(alphabet, p_num)
    except _TableSizeException:
        length, _ = _get_table_size(_DEFAULT_ALPHABET, p_num)

    return length



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides a pretty printer filter that generates a todo string based on a format
string.
"""

from topydo.lib.ListFormat import ListFormatParser
from topydo.lib.PrettyPrinterFilter import PrettyPrinterFilter


class PrettyPrinterFormatFilter(PrettyPrinterFilter):
    def __init__(self, p_todolist, p_format=None):
        super().__init__()
        self.parser = ListFormatParser(p_todolist, p_format)

    def filter(self, p_todo_str, p_todo):
        p_todo_str = self.parser.parse(p_todo)

        return p_todo_str



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import commonprefix

import urwid

from topydo.ui.columns.CompletionBoxWidget import CompletionBoxWidget


class CommandLineWidget(urwid.Edit):
    def __init__(self, p_completer, *args, **kwargs):

        self.history = []
        self.history_pos = None
        # temporary history storage for edits before cmd execution
        self.history_tmp = []

        self.completer = p_completer
        self.completion_box = CompletionBoxWidget()
        self._surrounding_text = None  # text before insertion of completion

        super().__init__(*args, **kwargs)
        urwid.register_signal(CommandLineWidget, ['blur',
                                                  'execute_command',
                                                  'show_completions',
                                                  'hide_completions'])

    def clear(self):
        self.set_edit_text("")

    def _blur(self):
        self.clear()
        urwid.emit_signal(self, 'blur')

    def _emit_command(self):
        if len(self.edit_text) > 0:
            urwid.emit_signal(self, 'execute_command', self.edit_text)
            self._save_to_history()
            self.clear()

    def _save_to_history(self):
        if len(self.edit_text) > 0:
            self.history.append(self.edit_text)

        self.history_pos = len(self.history)
        self.history_tmp = self.history[:] # sync temporary storage with real history
        self.history_tmp.append('')

    def _history_move(self, p_step):
        """
        Changes current value of the command-line to the value obtained from
        history_tmp list with index calculated by addition of p_step to the
        current position in the command history (history_pos attribute).

        Also saves value of the command-line (before changing it) to history_tmp
        for potential later access.
        """
        if len(self.history) > 0:
            # don't pollute real history - use temporary storage
            self.history_tmp[self.history_pos] = self.edit_text
            self.history_pos = self.history_pos + p_step
            self.set_edit_text(self.history_tmp[self.history_pos])

    def _history_next(self):
        if self.history_pos != len(self.history):
            self._history_move(1)

    def _history_back(self):
        if self.history_pos != 0:
            self._history_move(-1)

    def insert_completion(self, p_insert):
        """
        Inserts currently chosen completion (p_insert parameter) into proper
        place in edit_text and adjusts cursor position accordingly.
        """
        start, end = self._surrounding_text
        final_text = start + p_insert + end
        self.set_edit_text(final_text)
        self.set_edit_pos(len(start) + len(p_insert))

    @property
    def completion_mode(self):
        return len(self.completion_box) > 1

    @completion_mode.setter
    def completion_mode(self, p_enable):
        if p_enable is True:
            urwid.emit_signal(self, 'show_completions')
        elif p_enable is False:
            self._surrounding_text = None
            if self.completion_mode:
                self.completion_box.clear()
                urwid.emit_signal(self, 'hide_completions')

    def _complete(self):
        """
        Main completion function.

        Gets list of potential completion candidates for currently edited word,
        completes it to the longest common part, and shows convenient completion
        widget (if multiple completions are returned) with currently selected
        candidate highlighted.
        """
        def find_word_start(p_text, p_pos):
            """ Returns position of the beginning of a word ending in p_pos. """
            return p_text.lstrip().rfind(' ', 0, p_pos) + 1

        def get_word_before_pos(p_text, p_pos):
            start = find_word_start(p_text, p_pos)

            return (p_text[start:p_pos], start)

        pos = self.edit_pos
        text = self.edit_text
        completer = self.completer

        word_before_cursor, start = get_word_before_pos(text, pos)
        completions = completer.get_completions(word_before_cursor, start == 0)
        # store slices before and after place for completion
        self._surrounding_text = (text[:start], text[pos:])

        single_completion = len(completions) == 1
        completion_done = single_completion and completions[0] == word_before_cursor

        if completion_done or not completions:
            self.completion_mode = False
            return
        elif single_completion:
            replacement = completions[0]
        else:
            replacement = commonprefix(completions)
            zero_candidate = replacement if replacement else word_before_cursor

            if zero_candidate != completions[0]:
                completions.insert(0, zero_candidate)

            self.completion_box.add_completions(completions)

        self.insert_completion(replacement)
        self.completion_mode = not single_completion

    def _completion_move(self, p_step, p_size):
        """
        Visually selects completion specified by p_step (positive numbers
        forwards, negative numbers backwards) and inserts it into edit_text.

        If p_step results in value out of range of currently evaluated
        completion candidates, list is rewinded to the start (if cycling
        forwards) or to the end (if cycling backwards).
        """
        current_position = self.completion_box.focus_position

        try:
            self.completion_box.set_focus(current_position + p_step)
        except IndexError:
            position = 0 if p_step > 0 else len(self.completion_box) - 1
            self.completion_box.set_focus(position)

        maxcols, = p_size
        size = (maxcols, self.completion_box.height)
        self.completion_box.calculate_visible(size)

        candidate = self.completion_box.focus.original_widget.text
        self.insert_completion(candidate)

    def _prev_completion(self, p_size):
        if self.completion_mode:
            self._completion_move(-1, p_size)

    def _next_completion(self, p_size):
        self._completion_move(1, p_size)

    def _home(self):
        """ Moves cursor to the beginning of the line. """
        self.set_edit_pos(0)

    def _end(self):
        """ Moves cursor to the end of the line. """
        end = len(self.edit_text)
        self.set_edit_pos(end)

    def _home_del(self):
        """ Deletes the line content before the cursor  """
        text = self.edit_text[self.edit_pos:]
        self.set_edit_text(text)
        self._home()

    def _end_del(self):
        """ Deletes the line content after the cursor  """
        text = self.edit_text[:self.edit_pos]
        self.set_edit_text(text)

    def keypress(self, p_size, p_key):
        tab_handler = self._complete

        if self.completion_mode:
            tab_handler = lambda: self._next_completion(p_size)

            if p_key not in {'tab', 'shift tab'}:
                self.completion_mode = False

        dispatch = {
            'enter': self._emit_command,
            'esc': self._blur,
            'up': self._history_back,
            'down': self._history_next,
            'ctrl a': self._home,
            'ctrl e': self._end,
            'ctrl u': self._home_del,
            'ctrl k': self._end_del,
            'tab': tab_handler,
            'shift tab': lambda: self._prev_completion(p_size)
        }

        try:
            dispatch[p_key]()
        except KeyError:
            super().keypress(p_size, p_key)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import Command
from topydo.lib.Config import config
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.TodoListBase import InvalidTodoException


class WriteCommand(Command):

    def postprocess_input_todo(self, p_todo):
        """
        Post-processes a parsed todo when adding it to the list.

        * It converts relative dates to absolute ones.
        * Automatically inserts a creation date if not present.
        * Handles more user-friendly dependencies with before:, partof: and
        after: tags
        """
        def convert_date(p_tag):
            value = p_todo.tag_value(p_tag)

            if value:
                dateobj = relative_date_to_date(value)
                if dateobj:
                    p_todo.set_tag(p_tag, dateobj.isoformat())

        def add_dependencies(p_tag):
            for value in p_todo.tag_values(p_tag):
                try:
                    dep = self.todolist.todo(value)

                    if p_tag == 'after':
                        self.todolist.add_dependency(p_todo, dep)
                    elif p_tag == 'before' or p_tag == 'partof':
                        self.todolist.add_dependency(dep, p_todo)
                    elif p_tag.startswith('parent'):
                        for parent in self.todolist.parents(dep):
                            self.todolist.add_dependency(parent, p_todo)
                    elif p_tag.startswith('child'):
                        for child in self.todolist.children(dep):
                            self.todolist.add_dependency(p_todo, child)
                except InvalidTodoException:
                    pass

                p_todo.remove_tag(p_tag, value)

        convert_date(config().tag_start())
        convert_date(config().tag_due())

        keywords = [
            'after',
            'before',
            'child-of',
            'childof',
            'children-of',
            'childrenof',
            'parent-of',
            'parentof',
            'parents-of',
            'parentsof',
            'partof',
        ]

        for keyword in keywords:
            add_dependencies(keyword)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from topydo.lib.MultiCommand import MultiCommand
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.Utils import is_valid_priority


class PriorityCommand(MultiCommand):
    def __init__(self, p_args, p_todolist,  # pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.last_argument = True
        self.delete = False

    def _execute_multi_specific(self):
        def normalize_priority(p_priority):
            match = re.search(r'\b([A-Z])\b', p_priority.upper())
            return match.group(1) if match else p_priority

        priority = None if self.delete else normalize_priority(self.args[-1])
        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))

        if is_valid_priority(priority):
            for todo in self.todos:
                old_priority = todo.priority()
                self.todolist.set_priority(todo, priority)

                if old_priority and priority and old_priority != priority:
                    self.out("Priority changed from {} to {}".format(
                        old_priority, priority))
                elif not old_priority:
                    self.out("Priority set to {}.".format(priority))

                self.out(self.printer.print_todo(todo))
        elif priority is None:
            for todo in self.todos:
                old_priority = todo.priority()
                self.todolist.set_priority(todo, None)

                if old_priority:
                    self.out("Priority removed.")

                self.out(self.printer.print_todo(todo))
        else:
            self.error("Invalid priority given.")

    def get_flags(self):
        return ("d", [])

    def process_flag(self, p_option, p_value):
        if p_option == '-d':
            self.delete = True
            self.last_argument = False
        else:
            raise NotImplementedError

    def usage(self):
        return """\
Synopsis: pri [-d] <NUMBER 1> [<NUMBER 2> ...] <PRIORITY>
          pri [-d] [-x] -e <EXPRESSION>\
"""

    def help(self):
        return """\
Sets the priority of todo(s) the given NUMBER(s) to the given PRIORITY.

Use the -d flag to remove the priority. It is also possible to prioritize items
with an EXPRESSION using the -e flag.  Use -x to also process todo items that
are normally invisible (as with the 'ls' subcommand).\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.DCommand import DCommand


class DeleteCommand(DCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

    def prompt_text(self):
        return "Also remove subtasks? [y/N] "

    def prefix(self):
        return "Removed: "

    def execute_specific_core(self, p_todo):
        self.todolist.delete(p_todo)

    def execute_specific(self, p_todo):
        self.out(self.prefix() + self.printer.print_todo(p_todo))
        self.execute_specific_core(p_todo)

    def usage(self):
        return """\
Synopsis: del [-f] <NUMBER 1> [<NUMBER 2> ...]
          del [-x] -e <EXPRESSION>\
"""

    def help(self):
        return """\
Deletes the todo item(s) with the given number(s) from the list.

It is also possible to delete items that match EXPRESSION using the -e flag.
Use -x to also process todo items that are normally invisible (as with the 'ls'
subcommand).\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from topydo.lib.MultiCommand import MultiCommand
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.printers.PrettyPrinter import PrettyPrinter


class DCommand(MultiCommand):
    """
    A common class for the 'do' and 'del' operations, because they're quite
    alike.
    """

    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.force = False
        self._delta = []
        self.condition = lambda _: True
        self.condition_failed_text = ""

    def get_flags(self):
        return ("f", ["force"])

    def process_flag(self, p_option, p_value):
        if p_option == "-f" or p_option == "--force":
            self.force = True

    def _uncompleted_children(self, p_todo):
        return sorted(
            [t for t in self.todolist.children(p_todo) if not t.is_completed()],
            key=self.todolist.number
        )

    def _print_list(self, p_todos):
        printer = PrettyPrinter()
        printer.add_filter(PrettyPrinterNumbers(self.todolist))
        self.out(printer.print_list(p_todos))

    def prompt_text(self):
        raise NotImplementedError

    def prefix(self):
        raise NotImplementedError

    def _process_subtasks(self, p_todo):
        children = self._uncompleted_children(p_todo)
        if children:
            self._print_list(children)

            if not self.force:
                confirmation = self.prompt(self.prompt_text())

            if not self.force and re.match('^y(es)?$', confirmation, re.I):
                for child in children:
                    self.execute_specific_core(child)
                    self.out(self.prefix() + self.printer.print_todo(child))

    def _print_unlocked_todos(self):
        if self._delta:
            self.out("The following todo item(s) became active:")
            self._print_list(self._delta)

    def _active_todos(self):
        """
        Returns a list of active todos, taking uncompleted subtodos into
        account.

        The stored length of the todolist is taken into account, to prevent new
        todos created by recurrence to pop up as newly activated tasks.
        Since these todos pop up at the end of the list, we cut off the list
        just before that point.
        """
        return [todo for todo in self.todolist.todos()
                if not self._uncompleted_children(todo) and todo.is_active()]

    def execute_specific(self, _):
        raise NotImplementedError

    def execute_specific_core(self, p_todo):
        """
        The core operation on the todo itself. Also used to operate on
        child/parent tasks.
        """
        raise NotImplementedError

    def _execute_multi_specific(self):
        old_active = self._active_todos()

        for todo in self.todos:
            if todo and self.condition(todo):
                self._process_subtasks(todo)
                self.execute_specific(todo)
            else:
                self.error(self.condition_failed_text)

        current_active = self._active_todos()
        self._delta = [todo for todo in current_active
                       if todo not in old_active]

    def execute_post_archive_actions(self):
        self._print_unlocked_todos()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.MultiCommand import MultiCommand


class Transaction(object):
    """
    This class implements basic handling of executing any subcommand on multiple
    todo items.
    """
    def __init__(self, p_subcommand=None, p_env_args=(), p_todo_ids=None):
        self._multi = issubclass(p_subcommand, MultiCommand)
        self._cmd = lambda op: p_subcommand(op, *p_env_args)
        self._todo_ids = p_todo_ids
        self._operations = []
        self._post_archive_actions = []
        self._cmd_name = p_subcommand.name()
        self.label = []

    def prepare(self, p_args):
        """
        Prepares list of operations to execute based on p_args, list of
        todo items contained in _todo_ids attribute and _subcommand
        attribute.
        """
        if self._todo_ids:
            id_position = p_args.index('{}')

            # Not using MultiCommand abilities would make EditCommand awkward
            if self._multi:
                p_args[id_position:id_position + 1] = self._todo_ids
                self._operations.append(p_args)
            else:
                for todo_id in self._todo_ids:
                    operation_args = p_args[:]
                    operation_args[id_position] = todo_id
                    self._operations.append(operation_args)
        else:
            self._operations.append(p_args)

        self._create_label()

    def _create_label(self):
        if len(self._operations) > 1:
            for operation in self._operations:
                self.label.append(self._cmd_name + ' ' +
                                  ' '.join(operation) + ';')
        else:
            self.label.append(self._cmd_name + ' ' +
                              ' '.join(self._operations[0]))

    def execute(self):
        """
        Executes each operation from _operations attribute.
        """
        last_operation = len(self._operations) - 1
        for i, operation in enumerate(self._operations):
            command = self._cmd(operation)

            if command.execute() is False:
                return False
            else:
                action = command.execute_post_archive_actions
                self._post_archive_actions.append(action)

                if i == last_operation:
                    return True

    def execute_post_archive_actions(self):
        for action in self._post_archive_actions:
            action()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides the Todo class.
"""

from datetime import date

from topydo.lib.Config import config
from topydo.lib.TodoBase import TodoBase
from topydo.lib.Utils import date_string_to_date


class Todo(TodoBase):
    """
    This class adds common functionality with respect to dates to the Todo
    base class, mainly by interpreting the start and due dates of task.
    """

    def __init__(self, p_str):
        TodoBase.__init__(self, p_str)
        self.attributes = {}

    def get_date(self, p_tag):
        """ Given a date tag, return a date object. """
        string = self.tag_value(p_tag)
        result = None

        try:
            result = date_string_to_date(string) if string else None
        except ValueError:
            pass

        return result

    def start_date(self):
        """ Returns a date object of the todo's start date. """
        return self.get_date(config().tag_start())

    def due_date(self):
        """ Returns a date object of the todo's due date. """
        return self.get_date(config().tag_due())

    def is_active(self):
        """
        Returns True when the start date is today or in the past and the
        task has not yet been completed.
        """
        start = self.start_date()
        return not self.is_completed() and (not start or start <= date.today())

    def is_overdue(self):
        """
        Returns True when the due date is in the past and the task has not
        yet been completed.
        """
        return not self.is_completed() and self.days_till_due() < 0

    def days_till_due(self):
        """
        Returns the number of days till the due date. Returns a negative number
        of days when the due date is in the past.
        Returns 0 when the task has no due date.
        """
        due = self.due_date()
        if due:
            diff = due - date.today()
            return diff.days
        return 0

    def length(self):
        """
        Returns the length (in days) of the task, by considering the start date
        and the due date. When there is no start date, its creation date is
        used. Returns 0 when one of these dates is missing.
        """
        start = self.start_date() or self.creation_date()
        due = self.due_date()

        if start and due and start < due:
            diff = due - start
            return diff.days
        else:
            return 0



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Utilities for formatting output with "list_format" option."""

import re

import arrow

from topydo.lib.Config import config
from topydo.lib.ProgressColor import progress_color
from topydo.lib.Utils import escape_ansi, get_terminal_size, humanize_date

MAIN_PATTERN = (r'^({{(?P<before>.+?)}})?'
                r'(?P<placeholder>{ph}|\[{ph}\])'
                r'({{(?P<after>.+?)}})?'
                r'(?P<whitespace> *)')

def _columns():
    """ Returns the number of columns of the terminal. """
    return get_terminal_size().columns

def _filler(p_str, p_len):
    """
    Returns p_str preceded by additional spaces if p_str is shorter than p_len.
    """
    to_fill = p_len - len(p_str)
    return to_fill*' ' + p_str

def humanize_dates(p_due=None, p_start=None, p_creation=None):
    """
    Returns string with humanized versions of p_due, p_start and p_creation.
    Examples:
    - all dates: "16 days ago, due in a month, started 2 days ago"
    - p_due and p_start: "due in a month, started 2 days ago"
    - p_creation and p_due: "16 days ago, due in a month"
    """
    dates_list = []
    if p_creation:
        dates_list.append(humanize_date(p_creation))
    if p_due:
        dates_list.append('due ' + humanize_date(p_due))
    if p_start:
        now = arrow.now().date()
        dates_list.append('{} {}'.format(
            'started' if p_start <= now else 'starts',
            humanize_date(p_start)
        ))

    return ', '.join(dates_list)

def _strip_placeholder_braces(p_matchobj):
    """
    Returns string with conditional braces around placeholder stripped and
    percent sign glued into placeholder character.

    Returned string is composed from 'start', 'before', 'placeholder', 'after',
    'whitespace', and 'end' match-groups of p_matchobj. Conditional braces are
    stripped from 'before' and 'after' groups. 'whitespace', 'start', and 'end'
    groups are preserved without any change.

    Using this function as an 'repl' argument in re.sub it is possible to turn:
        %{(}B{)}
    into:
        (%B)
    """
    before = p_matchobj.group('before') or ''
    placeholder = p_matchobj.group('placeholder')
    after = p_matchobj.group('after') or ''
    whitespace = p_matchobj.group('whitespace') or ''

    return before + '%' + placeholder + after + whitespace

def _unescape_percent_sign(p_str):
    """ Strips backslashes from escaped percent signs in p_str. """
    unescaped_str = re.sub(r'\\%', '%', p_str)

    return unescaped_str

def _remove_redundant_spaces(p_str):
    """ Removes spaces surrunding <TAB> character (\t) from p_str. """
    clean_str = re.sub(' *\t *', '\t', p_str)

    return clean_str

def _truncate(p_str, p_repl):
    """
    Returns p_str with truncated and ended with '...' version of p_repl.

    Place of the truncation is calculated depending on p_max_width.
    """
    # 4 is for '...' and an extra space at the end
    text_lim = _columns() - len(escape_ansi(p_str)) - 4
    truncated_str = re.sub(re.escape(p_repl), p_repl[:text_lim] + '...', p_str)

    return truncated_str

def _right_align(p_str):
    """
    Returns p_str with content after <TAB> character aligned right.

    Right alignment is done using proper number of spaces calculated from
    'line_width' attribute.
    """
    to_fill = _columns() - len(escape_ansi(p_str))

    if to_fill > 0:
        p_str = re.sub('\t', ' '*to_fill, p_str)
    else:
        p_str = re.sub('\t', ' ', p_str)

    return p_str

def color_block(p_todo):
    return '{} {}'.format(
        progress_color(p_todo).as_ansi(p_background=True),
        config().priority_color(p_todo.priority()).as_ansi(),
    )


class ListFormatError(Exception):
    pass


class ListFormatParser(object):
    """ Parser of format string. """
    def __init__(self, p_todolist, p_format=None):
        self.format_string = re.sub(r'\\t', '\t', p_format or config().list_format())
        self.todolist = p_todolist
        self.one_line = False
        self.placeholders = {
            # absolute creation date
            'c': lambda t: t.creation_date().isoformat() if t.creation_date() else '',

            # relative creation date
            'C': lambda t: humanize_date(t.creation_date()) if t.creation_date() else '',

            # absolute due date
            'd': lambda t: t.due_date().isoformat() if t.due_date() else '',

            # relative due date
            'D': lambda t: humanize_date(t.due_date()) if t.due_date() else '',

            # relative dates:  due, start
            'h': lambda t: humanize_dates(t.due_date(), t.start_date()),

            # relative dates in form:  creation, due, start
            'H': lambda t: humanize_dates(t.due_date(), t.start_date(), t.creation_date()),

            # todo ID
            'i': lambda t: str(self.todolist.number(t)),

            # todo ID, padded with spaces
            'I': lambda t: _filler(str(self.todolist.number(t)),
                self.todolist.max_id_length()),

            # list of tags (spaces) without hidden ones and due: and t:
            'k': lambda t: ' '.join([u'{}:{}'.format(tag, value)
                                     for tag, value in sorted(t.tags()) if
                                     tag not in config().hidden_tags() +
                                     [config().tag_start(), config().tag_due()]]),

            # list of all tags (spaces)
            'K': lambda t: ' '.join([u'{}:{}'.format(tag, value)
                                     for tag, value in sorted(t.tags())]),

            # line number
            'n': lambda t: str(self.todolist.linenumber(t)),

            # line number, padded with spaces
            'N': lambda t: _filler(str(self.todolist.linenumber(t)),
                self.todolist.max_id_length()),

            # priority
            'p': lambda t: t.priority() if t.priority() else '',

            # priority (or placeholder space)
            'P': lambda t: t.priority() if t.priority() else ' ',

            # raw text
            'r': lambda t: t.source(),

            # text
            's': lambda t: t.text(),

            # text (truncated if necessary)
            'S': lambda t: t.text(),

            # absolute start date
            't': lambda t: t.start_date().isoformat() if t.start_date() else '',

            # relative start date
            'T': lambda t: humanize_date(t.start_date()) if t.start_date() else '',

            # unique text ID
            'u': self.todolist.uid if self.todolist else lambda _ : '',

            # unique text ID, padded with spaces
            'U': lambda t: _filler(self.todolist.uid(t),
                self.todolist.max_id_length()),

            # absolute completion date
            'x': lambda t: 'x ' + t.completion_date().isoformat() if t.is_completed() else '',

            # relative completion date
            'X': lambda t: 'x ' + humanize_date(t.completion_date()) if t.is_completed() else '',

            'z': lambda t: color_block(t) if config().colors() else ' ',
        }
        self.format_list = self._preprocess_format()

    def _preprocess_format(self):
        """
        Preprocess the format_string attribute.

        Splits the format string on each placeholder and returns a list of
        tuples containing substring, placeholder name, and function
        retrieving content for placeholder (getter).

        Relevant placeholder functions (getters) are taken from
        'placeholders' attribute which is a dict. If no matching placeholder
        is found in 'placeholders' getter is set to None. Getter and
        placeholder are also always set to None in first element of the
        returned list, because it never contain a real placeholder (read
        re.split documentation for further information).
        """
        format_split = re.split(r'(?<!\\)%', self.format_string)
        preprocessed_format = []

        for idx, substr in enumerate(format_split):
            if idx == 0:
                getter = None
                placeholder = None
            else:
                pattern = MAIN_PATTERN.format(ph=r'\S')
                try:
                    placeholder = re.match(pattern, substr).group('placeholder').strip('[]')
                except AttributeError:
                    placeholder = None

                if placeholder == 'S':
                    self.one_line = True

                try:
                    getter = self.placeholders[placeholder]
                except KeyError:
                    getter = None
                    substr = re.sub(pattern, '', substr)

            format_elem = (substr, placeholder, getter)
            preprocessed_format.append(format_elem)

        return preprocessed_format

    def parse(self, p_todo):
        """
        Returns fully parsed string from 'format_string' attribute with all
        placeholders properly substituted by content obtained from p_todo.

        It uses preprocessed form of 'format_string' (result of
        ListFormatParser._preprocess_format) stored in 'format_list'
        attribute.
        """
        parsed_list = []
        repl_trunc = None

        for substr, placeholder, getter in self.format_list:
            repl = getter(p_todo) if getter else ''
            pattern = MAIN_PATTERN.format(ph=placeholder)

            if placeholder == 'S':
                repl_trunc = repl

            try:
                if repl == '':
                    substr = re.sub(pattern, '', substr)
                else:
                    substr = re.sub(pattern, _strip_placeholder_braces, substr)
                    substr = re.sub(r'(?<!\\)%({ph}|\[{ph}\])'.format(ph=placeholder), repl, substr)
            except re.error:
                raise ListFormatError

            parsed_list.append(substr)

        parsed_str = _unescape_percent_sign(''.join(parsed_list))
        parsed_str = _remove_redundant_spaces(parsed_str)

        if self.one_line and len(escape_ansi(parsed_str)) >= _columns():
            parsed_str = _truncate(parsed_str, repl_trunc)

        if re.search('.*\t', parsed_str):
            parsed_str = _right_align(parsed_str)

        return parsed_str.rstrip()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib import Filter
from topydo.lib.Command import Command
from topydo.lib.Config import config
from topydo.lib.Sorter import Sorter
from topydo.lib.View import View


class ExpressionCommand(Command):
    """
    A common class for commands operating on todos selected by expressions.
    """

    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.sort_expression = config().sort_string()
        self.show_all = False
        self.limit = config().list_limit()
        # Commands using last argument differently (i.e as something other than
        # todo ID/expression) have to set attribute below to True.
        self.last_argument = False

    def _filters(self):
        filters = []

        if not self.show_all:
            filters.append(Filter.DependencyFilter(self.todolist))
            filters.append(Filter.RelevanceFilter())

        args = self.args[:-1] if self.last_argument else self.args
        filters += Filter.get_filter_list(args)

        if not self.show_all:
            filters.append(Filter.LimitFilter(self.limit))

        return filters

    def _view(self):
        sorter = Sorter(self.sort_expression)
        filters = self._filters()

        return View(sorter, filters, self.todolist)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the parse function which parses a single line of a
todo.txt file.
"""

import re

from topydo.lib.Utils import date_string_to_date

_DATE_MATCH = r'\d{4}-\d{2}-\d{2}'

_COMPLETED_HEAD_MATCH = re.compile(
    r'x ((?P<completionDate>' + _DATE_MATCH + ') )?' + '((?P<creationDate>' +
    _DATE_MATCH + ') )?(?P<rest>.*)')

_NORMAL_HEAD_MATCH = re.compile(
    r'(\((?P<priority>[A-Z])\) )?' + '((?P<creationDate>' + _DATE_MATCH +
    ') )?(?P<rest>.*)')

_TAG_MATCH = re.compile(r'(?![0-9+]{1,2}:[0-9]{1,2}$)(?P<tag>[^:]+):(?P<value>.+)')
_PROJECT_MATCH = re.compile(r'\+(\S*\w)')
_CONTEXT_MATCH = re.compile(r'@(\S*\w)')


def parse_line(p_string):
    """
    Parses a single line as can be encountered in a todo.txt file.
    First checks whether the standard elements are present, such as priority,
    creation date, completeness check and the completion date.

    Then the rest of the analyzed for any occurrences of contexts, projects or
    tags.

    Returns an dictionary with the default values as shown below.
    """
    result = {
        'completed': False,
        'completionDate': None,
        'priority': None,
        'creationDate': None,
        'text': "",
        'projects': [],
        'contexts': [],
        'tags': {},
    }

    completed_head = _COMPLETED_HEAD_MATCH.match(p_string)
    normal_head = _NORMAL_HEAD_MATCH.match(p_string)

    rest = p_string

    if completed_head:
        result['completed'] = True

        completion_date = completed_head.group('completionDate')
        try:
            result['completionDate'] = date_string_to_date(completion_date)
        except ValueError:
            pass

        creation_date = completed_head.group('creationDate')

        try:
            result['creationDate'] = date_string_to_date(creation_date)
        except ValueError:
            pass

        rest = completed_head.group('rest')
    elif normal_head:
        result['priority'] = normal_head.group('priority')

        creation_date = normal_head.group('creationDate')

        try:
            result['creationDate'] = date_string_to_date(creation_date)
        except ValueError:
            pass

        rest = normal_head.group('rest')

    for word in rest.split():
        project = _PROJECT_MATCH.match(word)
        if project:
            result['projects'].append(project.group(1))

        context = _CONTEXT_MATCH.match(word)
        if context:
            result['contexts'].append(context.group(1))

        tag = _TAG_MATCH.match(word)
        if tag:
            tag_name = tag.group('tag')
            tag_value = tag.group('value')
            try:
                result['tags'][tag_name].append(tag_value)
            except KeyError:
                result['tags'][tag_name] = [tag_value]
        else:
            result['text'] += word + ' '

    # strip trailing space from resulting text
    result['text'] = result['text'][:-1]

    return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from topydo.lib.HashListValues import max_id_length
from topydo.lib.Utils import translate_key_to_config
from topydo.ui.columns.TodoWidget import TodoWidget
from topydo.ui.columns.Utils import PaletteItem


def get_execute_signal(p_prefix):
    if p_prefix == 'cmdv':
        signal = 'execute_command'
    else:
        signal = 'execute_command_silent'

    return signal


class TodoListWidget(urwid.LineBox):
    def __init__(self, p_view, p_title, p_keymap):
        self._view = None

        self.keymap = p_keymap
        # store a state for multi-key shortcuts (e.g. 'gg')
        self.keystate = None
        # store offset length for postpone command (e.g. '3' for 'p3w')
        self._pp_offset = None

        self._title = urwid.Text(p_title, align='center')
        self._title_widget = urwid.AttrMap(self._title, PaletteItem.DEFAULT)

        self.todolist = urwid.SimpleFocusListWalker([])
        self.listbox = urwid.ListBox(self.todolist)
        self.view = p_view

        pile = urwid.Pile([
            (1, urwid.Filler(self._title_widget)),
            (1, urwid.Filler(urwid.Divider('\u2500'))),
            ('weight', 1, self.listbox),
        ])

        pile.focus_position = 2

        super().__init__(pile)

        urwid.register_signal(TodoListWidget, ['execute_command_silent',
                                               'execute_command',
                                               'refresh',
                                               'add_pending_action',
                                               'remove_pending_action',
                                               'repeat_cmd',
                                               'column_action',
                                               'show_keystate',
                                               'toggle_mark',
                                               ])

    @property
    def view(self):
        return self._view

    @view.setter
    def view(self, p_view):
        self._view = p_view
        self.update()

    @property
    def title(self):
        return self._title.text

    @title.setter
    def title(self, p_title):
        self._title.set_text(p_title)

    def update(self):
        """
        Updates the todo list according to the todos in the view associated
        with this list.
        """
        old_focus_position = self.todolist.focus
        id_length = max_id_length(self.view.todolist.count())

        del self.todolist[:]

        for group, todos in self.view.groups.items():
            if len(self.view.groups) > 1:
                grouplabel = ", ".join(group)
                self.todolist.append(urwid.Text(grouplabel))
                self.todolist.append(urwid.Divider('-'))

            for todo in todos:
                todowidget = TodoWidget.create(todo, id_length)
                todowidget.number = self.view.todolist.number(todo)
                self.todolist.append(todowidget)
                self.todolist.append(urwid.Divider('-'))

        if old_focus_position:
            try:
                self.todolist.set_focus(old_focus_position)
            except IndexError:
                # scroll to the bottom if the last item disappeared from column
                # -2 for the same reason as in self._scroll_to_bottom()
                self.todolist.set_focus(len(self.todolist) - 2)

    def _go_down(self, p_size):
        self.listbox.keypress(p_size, 'down')
        self.listbox.set_focus_valign('bottom')

    def _scroll_to_top(self, p_size):
        if isinstance(self.todolist[0], urwid.Text):
            self.listbox.set_focus(2)
        else:
            self.listbox.set_focus(0)

        # see comment at _scroll_to_bottom
        self.listbox.calculate_visible(p_size)

    def _scroll_to_bottom(self, p_size):
        # -2 because the last Divider shouldn't be focused.
        end_pos = len(self.listbox.body) - 2
        self.listbox.set_focus(end_pos)

        # for some reason, set_focus doesn't rerender the list.
        # calculate_visible is the only public method (besides keypress) that
        # deals with pending focus changes.
        self.listbox.calculate_visible(p_size)

    @property
    def keystate(self):
        return self._keystate

    @keystate.setter
    def keystate(self, p_keystate):
        self._keystate = p_keystate
        keystate_to_show = p_keystate if p_keystate else ''
        urwid.emit_signal(self, 'show_keystate', keystate_to_show)

    def keypress(self, p_size, p_key):
        urwid.emit_signal(self, 'remove_pending_action')
        requires_further_input = ['postpone', 'postpone_s', 'pri']

        keymap, keystates = self.keymap

        shortcut = self.keystate or ''
        shortcut += translate_key_to_config(p_key)

        try:
            action = keymap[shortcut]
        except KeyError:
            action = None

        if action:
            if shortcut in keystates:
                # Supplied key-shortcut matches keystate and action. Save the
                # keystate in case user will hit another key and add an action
                # waiting for execution if user won't type anything further.
                self.keystate = shortcut
                if action not in requires_further_input:
                    self._add_pending_action(action, p_size)
            else:
                # Only action is matched. Handle it and reset keystate.
                self.resolve_action(action, p_size)
                self.keystate = None
            return
        else:
            if shortcut in keystates:
                self.keystate = shortcut
            else:
                try:
                    # Check whether current keystate matches built-in 'postpone'
                    # action.
                    mode = keymap[self.keystate]
                    if mode in ['postpone', 'postpone_s']:
                        if self._postpone_selected(p_key, mode) is not None:
                            self.keystate = None
                        else:
                            urwid.emit_signal(self, 'show_keystate',
                                              self.keystate + self._pp_offset)
                    else:
                        self.keystate = None
                    return
                except KeyError:
                    if not self.keystate:
                        # Single key that is not described in keymap config.
                        return self.listbox.keypress(p_size, p_key)
                    self.keystate = None
            return

    def mouse_event(self, p_size, p_event, p_button, p_column, p_row, p_focus):
        if p_event == 'mouse press':
            if p_button == 4:  # up
                self.listbox.keypress(p_size, 'up')
                return
            elif p_button == 5:  # down:
                self._go_down(p_size)
                return

        return super().mouse_event(p_size,  # pylint: disable=E1102
                                   p_event,
                                   p_button,
                                   p_column,
                                   p_row,
                                   p_focus)

    # pylint: disable=no-self-use
    def selectable(self):
        return True

    def _toggle_marked_status(self):
        try:
            todo = self.listbox.focus.todo
            todo_id = str(self.view.todolist.number(todo))
            if urwid.emit_signal(self, 'toggle_mark', todo_id):
                self.listbox.focus.mark()
            else:
                self.listbox.focus.unmark()
        except AttributeError:
            # No todo item selected
            pass

    def _mark_all(self):
        for todo in self.listbox.body:
            if isinstance(todo, TodoWidget):
                todo_id = str(self.view.todolist.number(todo.todo))
                urwid.emit_signal(self, 'toggle_mark', todo_id, 'mark')
                todo.mark()

    def _execute_on_selected(self, p_cmd_str, p_execute_signal):
        """
        Executes command specified by p_cmd_str on selected todo item.

        p_cmd_str should be a string with one replacement field ('{}') which
        will be substituted by id of the selected todo item.

        p_execute_signal is the signal name passed to the main loop. It should
        be one of 'execute_command' or 'execute_command_silent'.
        """
        try:
            todo = self.listbox.focus.todo
            todo_id = str(self.view.todolist.number(todo))

            urwid.emit_signal(self, p_execute_signal, p_cmd_str, todo_id)

            # force screen redraw after editing
            if p_cmd_str.startswith('edit'):
                urwid.emit_signal(self, 'refresh')
        except AttributeError:
            # No todo item selected
            pass

    def resolve_action(self, p_action_str, p_size=None):
        """
        Checks whether action specified in p_action_str is "built-in" or
        contains topydo command (i.e. starts with 'cmd') and forwards it to
        proper executing methods.

        p_size should be specified for some of the builtin actions like 'up' or
        'home' as they can interact with urwid.ListBox.keypress or
        urwid.ListBox.calculate_visible.
        """
        if p_action_str.startswith(('cmd ', 'cmdv ')):
            prefix, cmd = p_action_str.split(' ', 1)
            execute_signal = get_execute_signal(prefix)

            if '{}' in cmd:
                self._execute_on_selected(cmd, execute_signal)
            else:
                urwid.emit_signal(self, execute_signal, cmd)
        else:
            self.execute_builtin_action(p_action_str, p_size)

    def execute_builtin_action(self, p_action_str, p_size=None):
        """
        Executes built-in action specified in p_action_str.

        Currently supported actions are: 'up', 'down', 'home', 'end',
        'first_column', 'last_column', 'prev_column', 'next_column',
        'append_column', 'insert_column', 'edit_column', 'delete_column',
        'copy_column', swap_right', 'swap_left', 'postpone', 'postpone_s',
        'pri', 'mark', 'mark_all, 'reset' and 'repeat'.
        """
        column_actions = ['first_column',
                          'last_column',
                          'prev_column',
                          'next_column',
                          'append_column',
                          'insert_column',
                          'edit_column',
                          'delete_column',
                          'copy_column',
                          'swap_left',
                          'swap_right',
                          'reset',
                          ]

        if p_action_str in column_actions:
            urwid.emit_signal(self, 'column_action', p_action_str)
        elif p_action_str == 'up':
            self.listbox.keypress(p_size, p_action_str)
        elif p_action_str == 'down':
            self._go_down(p_size)
        elif p_action_str == 'home':
            self._scroll_to_top(p_size)
        elif p_action_str == 'end':
            self._scroll_to_bottom(p_size)
        elif p_action_str in ['postpone', 'postpone_s']:
            pass
        elif p_action_str == 'pri':
            pass
        elif p_action_str == 'mark':
            self._toggle_marked_status()
        elif p_action_str == 'mark_all':
            self._mark_all()
        elif p_action_str == 'repeat':
            self._repeat_cmd()

    def _add_pending_action(self, p_action, p_size):
        """
        Creates action waiting for execution and forwards it to the mainloop.
        """
        def generate_callback():
            def callback(*args):
                self.resolve_action(p_action, p_size)
                self.keystate = None

            return callback

        urwid.emit_signal(self, 'add_pending_action', generate_callback())

    def _postpone_selected(self, p_pattern, p_mode):
        """
        Postpones selected todo item by <COUNT><PERIOD>.

        Returns True after 'postpone' command is called (i.e. p_pattern is valid
        <PERIOD>), False when p_pattern is invalid and None if p_pattern is
        digit (i.e. part of <COUNT>).

        p_pattern accepts digit (<COUNT>) or one of the <PERIOD> letters:
        'd'(ay), 'w'(eek), 'm'(onth), 'y'(ear). If digit is specified, it is
        appended to _pp_offset attribute. If p_pattern contains one of the
        <PERIOD> letters, 'postpone' command is forwarded to execution with
        value of _pp_offset attribute used as <COUNT>. If _pp_offset is None,
        <COUNT> is set to 1.

        p_mode should be one of 'postpone_s' or 'postpone'. It decides whether
        'postpone' command should be called with or without '-s' flag.
        """
        if p_pattern.isdigit():
            if not self._pp_offset:
                self._pp_offset = ''
            self._pp_offset += p_pattern
            result = None
        else:
            if p_pattern in ['d', 'w', 'm', 'y', 'b']:
                offset = self._pp_offset or '1'
                if p_mode == 'postpone':
                    pp_cmd = 'cmd postpone {} '
                else:
                    pp_cmd = 'cmd postpone -s {} '
                pp_cmd += offset + p_pattern
                self.resolve_action(pp_cmd)
                result = True
            self._pp_offset = None
            result = False
        return result

    def _repeat_cmd(self):
        try:
            todo = self.listbox.focus.todo
            todo_id = str(self.view.todolist.number(todo))
        except AttributeError:
            todo_id = None

        urwid.emit_signal(self, 'repeat_cmd', todo_id)

    def highlight(self, p_highlight):
        if p_highlight:
            self._title_widget.set_attr_map({None: PaletteItem.DEFAULT_FOCUS})
        else:
            self._title_widget.set_attr_map({None: PaletteItem.DEFAULT})
